##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Example ideal parameter block for the VLE calucations for a
Benzene-Toluene-o-Xylene system.
"""

# Import Python libraries
import logging
import os

# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    Expression,
    log,
    NonNegativeReals,
    value,
    Var,
    Set,
    Param,
    sqrt,
    log10,
)
from pyomo.opt import SolverFactory, TerminationCondition
from pyomo.util.calc_var_value import calculate_variable_from_constraint

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType
)
from idaes.core.util.initialization import solve_indexed_blocks
from idaes.core.util.misc import add_object_reference
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_unfixed_variables,
)
from idaes.core.util.misc import extract_data
from idaes.dmf import DMF
from idaes.dmf.resource import Resource, TidyUnitData

# Set up logger
_log = logging.getLogger(__name__)

# Set up DMF (use working directory, wherever that is)
_dmf = DMF(".")


@declare_process_block_class("HDAParameterBlock")
class HDAParameterData(PhysicalParameterBlock):
    CONFIG = PhysicalParameterBlock.CONFIG()

    def _get_param(self, name: str):
        """Helper method to convert data stored in a DMF resource into a Pyomo rule.

        The intermediate form, :class:`TidyUnitData`, is constructed from the data
        stored in the JSON of the DMF resource.

        Args:
            name: Primary alias of DMF resource with stored data.

        Returns:
            function for `initialize` keyword of `Param` class constructor.

        Raises:
            KeyError: Resource not found
            ValueError: Data in resource is missing or malformed

        Side-effects:
            For future reference, the resource itself is stored in an instance dict
            called `self._property_data` where the keys are the parameter names and
            the values are instances of :class:`idaes.dmf.resource.Resource`.
        """
        r = _dmf.find_one(name=name)
        if not hasattr(r, "data"):
            raise KeyError(f"No resource named {name} was found")
        try:
            dframe = TidyUnitData(r.data)
        except ValueError as err:
            raise ValueError(f"While extracting data for {name}: {err}")
        self._property_resources[name] = r
        return extract_data(dframe.param_data)

    def get_property_resource(self, name: str) -> Resource:
        return self._property_resources[name]

    def build(self):
        """
        Callable method for Block construction.
        """
        super(HDAParameterData, self).build()

        self._property_resources = {}  # see _get_param(), get_property_resource()

        self.state_block_class = IdealStateBlock

        self.component_list = Set(
            initialize=["benzene", "toluene", "hydrogen", "methane"]
        )

        self.phase_list = Set(initialize=["Liq", "Vap"], ordered=True)

        # List of components in each phase (optional)
        self.phase_comp = {"Liq": self.component_list, "Vap": self.component_list}

        # List of phase equilibrium index
        self.phase_equilibrium_idx = Set(initialize=[1, 2, 3, 4, 5])

        self.phase_equilibrium_list = {
            1: ["benzene", ("Vap", "Liq")],
            2: ["toluene", ("Vap", "Liq")],
            3: ["hydrogen", ("Vap", "Liq")],
            4: ["methane", ("Vap", "Liq")],
            5: ["diphenyl", ("Vap", "Liq")],
        }

        # Thermodynamic reference state
        self.pressure_ref = Param(
            mutable=True, default=101325, doc="Reference pressure [Pa]"
        )
        self.temperature_ref = Param(
            mutable=True, default=298.15, doc="Reference temperature [K]"
        )

        self.pressure_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=self._get_param("pressure_crit"),
            doc="Critical pressure [Pa]",
        )

        self.temperature_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=self._get_param("temperature_crit"),
            doc="Critical temperature [K]",
        )

        # Gas Constant
        self.gas_const = Param(
            within=NonNegativeReals,
            mutable=False,
            default=8.314,
            doc="Gas Constant [J/mol.K]",
        )

        self.mw_comp = Param(
            self.component_list,
            mutable=False,
            initialize=self._get_param("mw"),
            doc="molecular weight Kg/mol",
        )

        self.dens_liq_params = Param(
            self.component_list,
            ["1", "2", "3", "4"],
            mutable=False,
            initialize=self._get_param("cp_liq"),
            doc="Parameters to compute liquid densities",
        )

        self.temperature_boil = Param(
            self.component_list,
            mutable=False,
            initialize=self._get_param("bp"),
            doc="Pure component boiling points at standard pressure [K]",
        )

        self.cp_ig = Param(
            self.phase_list,
            self.component_list,
            ["1", "2", "3", "4", "5"],
            mutable=False,
            initialize=self._get_param("cp_ig"),
            doc="parameters to compute Cp_comp",
        )

        self.pressure_sat_coeff = Param(
            self.component_list,
            ["A", "B", "C"],
            mutable=False,
            initialize=self._get_param("pressure_sat_coeff"),
            doc="parameters to compute Cp_comp",
        )

        self.dh_vap = Param(
            self.component_list,
            mutable=False,
            initialize=self._get_param("dh_vap"),
            doc="heat of vaporization",
        )

    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.add_properties(
            {
                "flow_mol": {"method": None, "units": "mol/s"},
                "flow_mol_phase_comp": {"method": None, "units": "mol/s"},
                "mole_frac": {"method": None, "units": "none"},
                "temperature": {"method": None, "units": "K"},
                "pressure": {"method": None, "units": "Pa"},
                "flow_mol_phase": {"method": None, "units": "mol/s"},
                "dens_mol_phase": {"method": "_dens_mol_phase", "units": "mol/m^3"},
                "pressure_sat": {"method": "_pressure_sat", "units": "Pa"},
                "mole_frac_phase": {"method": "_mole_frac_phase", "units": "no unit"},
                "enth_mol_phase_comp": {
                    "method": "_enth_mol_phase_comp",
                    "units": "J/mol",
                },
                "enth_mol_phase": {"method": "_enth_mol_phase", "units": "J/mol"},
                "entr_mol_phase_comp": {
                    "method": "_entr_mol_phase_comp",
                    "units": "J/mol",
                },
                "entr_mol_phase": {"method": "_entr_mol_phase", "units": "J/mol"},
                "temperature_bubble": {"method": "_temperature_bubble", "units": "K"},
                "temperature_dew": {"method": "_temperature_dew", "units": "K"},
                "pressure_bubble": {"method": "_pressure_bubble", "units": "Pa"},
                "pressure_dew": {"method": "_pressure_dew", "units": "Pa"},
                "fug_vap": {"method": "_fug_vap", "units": "Pa"},
                "fug_liq": {"method": "_fug_liq", "units": "Pa"},
                "dh_vap": {"method": "_dh_vap", "units": "J/mol"},
                "ds_vap": {"method": "_ds_vap", "units": "J/mol.K"},
            }
        )

        obj.add_default_units(
            {
                "time": "s",
                "length": "m",
                "mass": "g",
                "amount": "mol",
                "temperature": "K",
                "energy": "J",
                "holdup": "mol",
            }
        )


class _IdealStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(
        blk,
        state_args=None,
        state_vars_fixed=False,
        hold_state=False,
        outlvl=1,
        solver="ipopt",
        optarg={"tol": 1e-8},
    ):
        """
        Initialization routine for property package.
        Keyword Arguments:
            state_args: Initial guess for state variables.
            outlvl : sets output level of initialization routine
                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = include solver output infomation (tee=True)
            optarg : solver options dictionary object (default=None)
            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed by the
                                       control volume 1D. Control volume 0D
                                       does not fix the state vars, so will
                                       be False if this state block is used
                                       with 0D blocks.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """

        _log.info("Starting {} initialization".format(blk.name))

        # Fix state variables if not already fixed
        if state_vars_fixed is False:
            Fflag = {}
            Pflag = {}
            Tflag = {}

            for k in blk.keys():
                for p in blk[k]._params.phase_list:
                    for j in blk[k]._params.component_list:
                        if blk[k].flow_mol_phase_comp[p, j].fixed is True:
                            Fflag[k, p, j] = True
                        else:
                            Fflag[k, p, j] = False
                            if state_args is None:
                                blk[k].flow_mol_phase_comp[p, j].fix(
                                    1 / len(blk[k]._params.component_list)
                                )
                            else:
                                blk[k].flow_mol_phase_comp[p, j].fix(
                                    state_args["flow_mol_phase_comp"][p, j]
                                )

                if blk[k].pressure.fixed is True:
                    Pflag[k] = True
                else:
                    Pflag[k] = False
                    if state_args is None:
                        blk[k].pressure.fix(101325.0)
                    else:
                        blk[k].pressure.fix(state_args["pressure"])

                if blk[k].temperature.fixed is True:
                    Tflag[k] = True
                else:
                    Tflag[k] = False
                    if state_args is None:
                        blk[k].temperature.fix(325)
                    else:
                        blk[k].temperature.fix(state_args["temperature"])

            # -----------------------------------------------------------------
            # If input block, return flags, else release state
            flags = {"Fflag": Fflag, "Pflag": Pflag, "Tflag": Tflag}

        else:
            # Check when the state vars are fixed already result in dof 0
            for k in blk.keys():
                if degrees_of_freedom(blk[k]) != 0:
                    raise Exception(
                        "State vars fixed but degrees of freedom "
                        "for state block is not zero during "
                        "initialization."
                    )
        # Set solver options
        if outlvl > 1:
            stee = True
        else:
            stee = False

        if optarg is None:
            sopt = {"tol": 1e-8}
        else:
            sopt = optarg

        opt = SolverFactory("ipopt")
        opt.options = sopt

        # ---------------------------------------------------------------------
        # If present, initialize bubble and dew point calculations
        for k in blk.keys():
            if hasattr(blk[k], "eq_temperature_dew"):
                calculate_variable_from_constraint(
                    blk[k].temperature_dew, blk[k].eq_temperature_dew
                )

            if hasattr(blk[k], "eq_pressure_dew"):
                calculate_variable_from_constraint(
                    blk[k].pressure_dew, blk[k].eq_pressure_dew
                )

        if outlvl > 0:
            _log.info(
                "Dew and bubble points initialization for "
                "{} completed".format(blk.name)
            )

        # ---------------------------------------------------------------------
        # If flash, initialize T1 and Teq
        for k in blk.keys():
            if blk[k].config.has_phase_equilibrium and not blk[k].config.defined_state:
                blk[k]._t1.value = max(
                    blk[k].temperature.value, blk[k].temperature_bubble.value
                )
                blk[k]._teq.value = min(blk[k]._t1.value, blk[k].temperature_dew.value)

        if outlvl > 0:
            _log.info(
                "Equilibrium temperature initialization for "
                "{} completed".format(blk.name)
            )

        # ---------------------------------------------------------------------
        # Initialize flow rates and compositions
        # TODO : This will need ot be generalised more when we move to a
        # modular implementation
        for k in blk.keys():
            # Deactivate equilibrium constraints, as state is fixed
            if hasattr(blk[k], "equilibrium_constraint"):
                blk[k].equilibrium_constraint.deactivate()

        free_vars = 0
        for k in blk.keys():
            free_vars += number_unfixed_variables(blk[k])
        if free_vars > 0:
            try:
                results = solve_indexed_blocks(opt, [blk], tee=stee)
            except:
                results = None
        else:
            results = None

        for k in blk.keys():
            # Reactivate equilibrium constraints
            if hasattr(blk[k], "equilibrium_constraint"):
                blk[k].equilibrium_constraint.activate()

        if outlvl > 0:
            if (
                results is None
                or results.solver.termination_condition == TerminationCondition.optimal
            ):
                _log.info(
                    "Property initialization for " "{} completed".format(blk.name)
                )
            else:
                _log.warning(
                    "Property initialization for " "{} failed".format(blk.name)
                )

        # ---------------------------------------------------------------------
        # Return state to initial conditions
        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                blk.release_state(flags)

        if outlvl > 0:
            _log.info("Initialization completed for {}".format(blk.name))

    def release_state(blk, flags, outlvl=0):
        """
        Method to relase state variables fixed during initialization.
        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        if flags is None:
            return

        # Unfix state variables
        for k in blk.keys():
            for p in blk[k]._params.phase_list:
                for j in blk[k]._params.component_list:
                    if flags["Fflag"][k, p, j] is False:
                        blk[k].flow_mol_phase_comp[p, j].unfix()
            if flags["Pflag"][k] is False:
                blk[k].pressure.unfix()
            if flags["Tflag"][k] is False:
                blk[k].temperature.unfix()

        if outlvl > 0:
            if outlvl > 0:
                _log.info("{} states released.".format(blk.name))


@declare_process_block_class("IdealStateBlock", block_class=_IdealStateBlock)
class IdealStateBlockData(StateBlockData):
    """An example property package for ideal VLE."""

    def build(self):
        """Callable method for Block construction."""
        super(IdealStateBlockData, self).build()

        # Add state variables
        self.flow_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            initialize=0.5,
            bounds=(1e-8, 100),
            doc="Phase-component molar flow rates [mol/s]",
        )

        self.pressure = Var(
            initialize=101325,
            bounds=(101325, 400000),
            domain=NonNegativeReals,
            doc="State pressure [Pa]",
        )
        self.temperature = Var(
            initialize=298.15,
            bounds=(298.15, 1000),
            domain=NonNegativeReals,
            doc="State temperature [K]",
        )

        # Add supporting variables
        def flow_mol_phase(b, p):
            return sum(b.flow_mol_phase_comp[p, j] for j in b._params.component_list)

        self.flow_mol_phase = Expression(
            self._params.phase_list,
            rule=flow_mol_phase,
            doc="Phase molar flow rates [mol/s]",
        )

        def flow_mol(b):
            return sum(
                b.flow_mol_phase_comp[p, j]
                for j in b._params.component_list
                for p in b._params.phase_list
            )

        self.flow_mol = Expression(rule=flow_mol, doc="Total molar flowrate [mol/s]")

        def mole_frac_phase(b, p, j):
            return b.flow_mol_phase_comp[p, j] / b.flow_mol_phase[p]

        self.mole_frac_phase = Expression(
            self._params.phase_list,
            self._params.component_list,
            rule=mole_frac_phase,
            doc="Phase mole fractions [-]",
        )

        def mole_frac(b, j):
            return (
                sum(b.flow_mol_phase_comp[p, j] for p in b._params.phase_list)
                / b.flow_mol
            )

        self.mole_frac = Expression(
            self._params.component_list,
            rule=mole_frac,
            doc="Mixture mole fractions [-]",
        )

        # Reaction Stoichiometry
        add_object_reference(
            self, "phase_equilibrium_list_ref", self._params.phase_equilibrium_list
        )

        if self.config.has_phase_equilibrium and self.config.defined_state is False:
            # Definition of equilibrium temperature for smooth VLE
            self._teq = Var(
                initialize=self.temperature.value,
                doc="Temperature for calculating phase equilibrium",
            )
            self._t1 = Var(
                initialize=self.temperature.value,
                doc="Intermediate temperature for calculating Teq",
            )

            self.eps_1 = Param(
                default=0.01, mutable=True, doc="Smoothing parameter for Teq"
            )
            self.eps_2 = Param(
                default=0.0005, mutable=True, doc="Smoothing parameter for Teq"
            )

            # PSE paper Eqn 13
            def rule_t1(b):
                return b._t1 == 0.5 * (
                    b.temperature
                    + b.temperature_bubble
                    + sqrt((b.temperature - b.temperature_bubble) ** 2 + b.eps_1 ** 2)
                )

            self._t1_constraint = Constraint(rule=rule_t1)

            # PSE paper Eqn 14
            # TODO : Add option for supercritical extension
            def rule_teq(b):
                return b._teq == 0.5 * (
                    b._t1
                    + b.temperature_dew
                    - sqrt((b._t1 - b.temperature_dew) ** 2 + b.eps_2 ** 2)
                )

            self._teq_constraint = Constraint(rule=rule_teq)

            def rule_tr_eq(b, i):
                return b._teq / b._params.temperature_crit[i]

            self._tr_eq = Expression(
                self._params.component_list,
                rule=rule_tr_eq,
                doc="Component reduced temperatures [-]",
            )

            def rule_equilibrium(b, i):
                return b.fug_vap[i] == b.fug_liq[i]

            self.equilibrium_constraint = Constraint(
                self._params.component_list, rule=rule_equilibrium
            )

    # -----------------------------------------------------------------------------
    # Property Methods
    def _dens_mol_phase(self):
        self.dens_mol_phase = Var(
            self._params.phase_list, initialize=1.0, doc="Molar density [mol/m^3]"
        )

        def rule_dens_mol_phase(b, p):
            if p == "Vap":
                return b._dens_mol_vap()
            else:
                return b._dens_mol_liq()

        self.eq_dens_mol_phase = Constraint(
            self._params.phase_list, rule=rule_dens_mol_phase
        )

    def _enth_mol_phase_comp(self):
        self.enth_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            initialize=7e5,
            doc="Phase-component molar specific enthalpies [J/mol]",
        )

        def rule_enth_mol_phase_comp(b, p, j):
            if p == "Vap":
                return b._enth_mol_comp_vap(j)
            else:
                return b._enth_mol_comp_liq(j)

        self.eq_enth_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_enth_mol_phase_comp,
        )

    def _enth_mol_phase(self):
        self.enth_mol_phase = Var(
            self._params.phase_list,
            initialize=7e5,
            doc="Phase molar specific enthalpies [J/mol]",
        )

        def rule_enth_mol_phase(b, p):
            return b.enth_mol_phase[p] == sum(
                b.enth_mol_phase_comp[p, i] * b.mole_frac_phase[p, i]
                for i in b._params.component_list
            )

        self.eq_enth_mol_phase = Constraint(
            self._params.phase_list, rule=rule_enth_mol_phase
        )

    def _entr_mol_phase_comp(self):
        self.entr_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            doc="Phase-component molar specific entropies [J/mol.K]",
        )

        def rule_entr_mol_phase_comp(b, p, j):
            if p == "Vap":
                return b._entr_mol_comp_vap(j)
            else:
                return b._entr_mol_comp_liq(j)

        self.eq_entr_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_entr_mol_phase_comp,
        )

    def _entr_mol_phase(self):
        self.entr_mol_phase = Var(
            self._params.phase_list, doc="Phase molar specific enthropies [J/mol.K]"
        )

        def rule_entr_mol_phase(b, p):
            return b.entr_mol_phase[p] == sum(
                b.entr_mol_phase_comp[p, i] * b.mole_frac_phase[p, i]
                for i in b._params.component_list
            )

        self.eq_entr_mol_phase = Constraint(
            self._params.phase_list, rule=rule_entr_mol_phase
        )

    # -----------------------------------------------------------------------------
    # General Methods
    def get_material_flow_terms(self, p, j):
        """Create material flow terms for control volume."""
        if j in self._params.component_list:
            return self.flow_mol_phase_comp[p, j]
        else:
            return 0

    def get_enthalpy_flow_terms(self, p):
        """Create enthalpy flow terms."""
        return self.flow_mol_phase[p] * self.enth_mol_phase[p]

    def get_material_density_terms(self, p, j):
        """Create material density terms."""
        if j in self._params.component_list:
            return self.dens_mol_phase[p] * self.mole_frac_phase[p, j]
        else:
            return 0

    def get_enthalpy_density_terms(self, p):
        """Create enthalpy density terms."""
        return self.dens_mol_phase[p] * self.enth_mol_phase[p]

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def define_state_vars(self):
        """Define state vars."""
        return {
            "flow_mol_phase_comp": self.flow_mol_phase_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }

    # Property package utility functions
    def calculate_bubble_point_temperature(self, clear_components=True):
        """"To compute the bubble point temperature of the mixture."""

        if hasattr(self, "eq_temperature_bubble"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(
            self.temperature_bubble, self.eq_temperature_bubble
        )

        return self.temperature_bubble.value

        if clear_components is True:
            self.del_component(self.eq_temperature_bubble)
            self.del_component(self._p_sat_bubbleT)
            self.del_component(self.temperature_bubble)

    def calculate_dew_point_temperature(self, clear_components=True):
        """"To compute the dew point temperature of the mixture."""

        if hasattr(self, "eq_temperature_dew"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(
            self.temperature_dew, self.eq_temperature_dew
        )

        return self.temperature_dew.value

        # Delete the var/constraint created in this method that are part of the
        # IdealStateBlock if the user desires
        if clear_components is True:
            self.del_component(self.eq_temperature_dew)
            self.del_component(self._p_sat_dewT)
            self.del_component(self.temperature_dew)

    def calculate_bubble_point_pressure(self, clear_components=True):
        """"To compute the bubble point pressure of the mixture."""

        if hasattr(self, "eq_pressure_bubble"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(
            self.pressure_bubble, self.eq_pressure_bubble
        )

        return self.pressure_bubble.value

        # Delete the var/constraint created in this method that are part of the
        # IdealStateBlock if the user desires
        if clear_components is True:
            self.del_component(self.eq_pressure_bubble)
            self.del_component(self._p_sat_bubbleP)
            self.del_component(self.pressure_bubble)

    def calculate_dew_point_pressure(self, clear_components=True):
        """"To compute the dew point pressure of the mixture."""

        if hasattr(self, "eq_pressure_dew"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(self.pressure_dew, self.eq_pressure_dew)

        return self.pressure_dew.value

        # Delete the var/constraint created in this method that are part of the
        # IdealStateBlock if the user desires
        if clear_components is True:
            self.del_component(self.eq_pressure_dew)
            self.del_component(self._p_sat_dewP)
            self.del_component(self.pressure_dew)

    # -----------------------------------------------------------------------------
    # Bubble and Dew Points
    # Ideal-Ideal properties allow for the simplifications below
    # Other methods require more complex equations with shadow compositions

    # For future work, propose the following:
    # Core class writes a set of constraints Phi_L_i == Phi_V_i
    # Phi_L_i and Phi_V_i make calls to submethods which add shadow compositions
    # as needed
    def _temperature_bubble(self):
        self.temperature_bubble = Param(
            initialize=33.0, doc="Bubble point temperature (K)"
        )

    def _temperature_dew(self):

        self.temperature_dew = Var(initialize=298.15, doc="Dew point temperature (K)")

        def rule_psat_dew(b, j):
            return 1e5 * 10 ** (
                b._params.pressure_sat_coeff[j, "A"]
                - b._params.pressure_sat_coeff[j, "B"]
                / (b.temperature_dew + b._params.pressure_sat_coeff[j, "C"])
            )

        try:
            # Try to build expression
            self._p_sat_dewT = Expression(
                self._params.component_list, rule=rule_psat_dew
            )

            def rule_temp_dew(b):
                return (
                    b.pressure
                    * sum(
                        b.mole_frac[i] / b._p_sat_dewT[i]
                        for i in ["toluene", "benzene"]
                    )
                    - 1
                    == 0
                )

            self.eq_temperature_dew = Constraint(rule=rule_temp_dew)
        except AttributeError:
            # If expression fails, clean up so that DAE can try again later
            # Deleting only var/expression as expression construction will fail
            # first; if it passes then constraint construction will not fail.
            self.del_component(self.temperature_dew)
            self.del_component(self._p_sat_dewT)

    def _pressure_bubble(self):
        self.pressure_bubble = Param(initialize=1e8, doc="Bubble point pressure (Pa)")

    def _pressure_dew(self):
        self.pressure_dew = Var(initialize=298.15, doc="Dew point pressure (Pa)")

        def rule_psat_dew(b, j):
            return 1e5 * 10 ** (
                b._params.pressure_sat_coeff[j, "A"]
                - b._params.pressure_sat_coeff[j, "B"]
                / (b.temperature + b._params.pressure_sat_coeff[j, "C"])
            )

        try:
            # Try to build expression
            self._p_sat_dewP = Expression(
                self._params.component_list, rule=rule_psat_dew
            )

            def rule_pressure_dew(b):
                return (
                    b.pressure_dew
                    * sum(
                        b.mole_frac[i] / b._p_sat_dewP[i]
                        for i in ["toluene", "benzene"]
                    )
                    - 1
                    == 0
                )

            self.eq_pressure_dew = Constraint(rule=rule_pressure_dew)
        except AttributeError:
            # If expression fails, clean up so that DAE can try again later
            # Deleting only var/expression as expression construction will fail
            # first; if it passes then constraint construction will not fail.
            self.del_component(self.pressure_dew)
            self.del_component(self._p_sat_dewP)

    # -----------------------------------------------------------------------------
    # Liquid phase properties
    def _dens_mol_liq(b):
        return b.dens_mol_phase["Liq"] == 1e3 * sum(
            b.mole_frac_phase["Liq", j]
            * b._params.dens_liq_params[j, "1"]
            / b._params.dens_liq_params[j, "2"]
            ** (
                1
                + (1 - b.temperature / b._params.dens_liq_params[j, "3"])
                ** b._params.dens_liq_params[j, "4"]
            )
            for j in ["benzene", "toluene"]
        )  # TODO: Need to include diphenyl here later

    def _fug_liq(self):
        def fug_liq_rule(b, i):
            if i in ["hydrogen", "methane"]:
                return b.mole_frac_phase["Liq", i]
            else:
                return b.pressure_sat[i] * b.mole_frac_phase["Liq", i]

        self.fug_liq = Expression(self._params.component_list, rule=fug_liq_rule)

    def _pressure_sat(self):
        self.pressure_sat = Var(
            self._params.component_list, initialize=101325, doc="Vapor pressure [Pa]"
        )

        def rule_P_sat(b, j):
            return (
                (log10(b.pressure_sat[j] * 1e-5) - b._params.pressure_sat_coeff[j, "A"])
                * (b._teq + b._params.pressure_sat_coeff[j, "C"])
            ) == -b._params.pressure_sat_coeff[j, "B"]

        self.eq_pressure_sat = Constraint(self._params.component_list, rule=rule_P_sat)

    def _enth_mol_comp_liq(b, j):
        return b.enth_mol_phase_comp["Liq", j] * 1e3 == (
            (b._params.cp_ig["Liq", j, "5"] / 5)
            * (b.temperature ** 5 - b._params.temperature_ref ** 5)
            + (b._params.cp_ig["Liq", j, "4"] / 4)
            * (b.temperature ** 4 - b._params.temperature_ref ** 4)
            + (b._params.cp_ig["Liq", j, "3"] / 3)
            * (b.temperature ** 3 - b._params.temperature_ref ** 3)
            + (b._params.cp_ig["Liq", j, "2"] / 2)
            * (b.temperature ** 2 - b._params.temperature_ref ** 2)
            + b._params.cp_ig["Liq", j, "1"]
            * (b.temperature - b._params.temperature_ref)
        )

    def _entr_mol_comp_liq(b, j):
        return b.entr_mol_phase_comp["Liq", j] * 1e3 == (
            (
                (b._params.cp_ig["Liq", j, "5"] / 4)
                * (b.temperature ** 4 - b._params.temperature_ref ** 4)
                + (b._params.cp_ig["Liq", j, "4"] / 3)
                * (b.temperature ** 3 - b._params.temperature_ref ** 3)
                + (b._params.cp_ig["Liq", j, "3"] / 2)
                * (b.temperature ** 2 - b._params.temperature_ref ** 2)
                + b._params.cp_ig["Liq", j, "2"]
                * (b.temperature - b._params.temperature_ref)
                + b._params.cp_ig["Liq", j, "1"]
                * log(b.temperature / b._params.temperature_ref)
            )
            - b._params.gas_const
            * log(b.mole_frac_phase["Liq", j] * b.pressure / b._params.pressure_ref)
        )

    # -----------------------------------------------------------------------------
    # Vapour phase properties
    def _dens_mol_vap(b):
        return b.pressure == (
            b.dens_mol_phase["Vap"] * b._params.gas_const * b.temperature
        )

    def _fug_vap(self):
        def fug_vap_rule(b, i):
            if i in ["hydrogen", "methane"]:
                return 1e-6
            else:
                return b.mole_frac_phase["Vap", i] * b.pressure

        self.fug_vap = Expression(self._params.component_list, rule=fug_vap_rule)

    def _dh_vap(self):
        # heat of vaporization
        add_object_reference(self, "dh_vap", self._params.dh_vap)

    def _ds_vap(self):
        # entropy of vaporization = dh_Vap/T_boil
        # TODO : something more rigorous would be nice
        self.ds_vap = Var(
            self._params.component_list,
            initialize=86,
            doc="Entropy of vaporization [J/mol.K]",
        )

        def rule_ds_vap(b, j):
            return b.dh_vap[j] == (b.ds_vap[j] * b._params.temperature_boil[j])

        self.eq_ds_vap = Constraint(self._params.component_list, rule=rule_ds_vap)

    def _enth_mol_comp_vap(b, j):
        return b.enth_mol_phase_comp["Vap", j] == b.dh_vap[j] + (
            (b._params.cp_ig["Vap", j, "5"] / 5)
            * (b.temperature ** 5 - b._params.temperature_ref ** 5)
            + (b._params.cp_ig["Vap", j, "4"] / 4)
            * (b.temperature ** 4 - b._params.temperature_ref ** 4)
            + (b._params.cp_ig["Vap", j, "3"] / 3)
            * (b.temperature ** 3 - b._params.temperature_ref ** 3)
            + (b._params.cp_ig["Vap", j, "2"] / 2)
            * (b.temperature ** 2 - b._params.temperature_ref ** 2)
            + b._params.cp_ig["Vap", j, "1"]
            * (b.temperature - b._params.temperature_ref)
        )

    def _entr_mol_comp_vap(b, j):
        return b.entr_mol_phase_comp["Vap", j] == (
            b.ds_vap[j]
            + (
                (b._params.cp_ig["Vap", j, "5"] / 4)
                * (b.temperature ** 4 - b._params.temperature_ref ** 4)
                + (b._params.cp_ig["Vap", j, "4"] / 3)
                * (b.temperature ** 3 - b._params.temperature_ref ** 3)
                + (b._params.cp_ig["Vap", j, "3"] / 2)
                * (b.temperature ** 2 - b._params.temperature_ref ** 2)
                + b._params.cp_ig["Vap", j, "2"]
                * (b.temperature - b._params.temperature_ref)
                + b._params.cp_ig["Vap", j, "1"]
                * log(b.temperature / b._params.temperature_ref)
            )
            - b._params.gas_const
            * log(b.mole_frac_phase["Vap", j] * b.pressure / b._params.pressure_ref)
        )



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
icon_mapping = {
    "cstr": "reactor_c.svg",
    "flash": "flash.svg",
    "gibbs_reactor": "reactor_g.svg",
    "heat_exchanger": "heat_exchanger_1.svg",
    "heater": "heater_2.svg",
    "heat_exchanger_1D": "heat_exchanger_1.svg",  # same as HeatExchanger
    "mixer": "mixer.svg",
    "plug_flow_reactor": "reactor_pfr.svg",
    "pressure_changer": "compressor.svg",
    "separator": "splitter.svg",
    "stoichiometric_reactor": "reactor_s.svg",
    "equilibrium_reactor": "reactor_e.svg",
    "feed": "feed.svg",
    "product": "product.svg",
    "feed_flash": "feed.svg",
    "statejunction": "NONE",
    "translator": "NONE",
    "packed_column": "packed_column_1.svg",
    "tray_column": "tray_column_1.svg",
    "default": "default.svg"
}


# TODO handle multi-icon mappings
#    "Heater": {"cooler.svg", "heater_1.svg", "heater_2.svg"},
#    "PressureChanger": {"compressor.svg", "expander.svg", "pump.svg", "fan.svg"},
# compressor and expander variations are identical when arrows are not present
#    "HeatExchanger": {"heat_exchanger_1.svg", "heat_exchanger_3.svg"},
# 1/2 and 3/4 are identical when arrows not present
#    "HeatExchanger1D": {"heat_exchanger_1.svg", "heat_exchanger_3.svg"},
#  same as HeadExchanger
#    "PackedColumn": {"packed_column_1.svg", "packed_column_2.svg",
#    "packed_column_3.svg", "packed_column_4.svg"},
#    "TrayColumn": {"tray_column_1.svg", "tray_column_2.svg", "tray_column_3.svg",
#    "tray_column_4.svg"},
#


# TODO need images, but the models don't exist yet
#    "StateJunction": "junction.svg", # small black circle "node"
#    "Translator": "translator.svg", # circle with a T in it
#    "Valve": {"valve_1.svg", "valve_2.svg", "valve_3.svg"}



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for idaes.dmf.commands
"""
import logging
import os
import shutil
import sys
import tempfile

#
import pytest

#
from idaes.dmf import dmfbase, commands, errors, workspace, util
from .util import init_logging

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def wspath():
    dirname = tempfile.mkdtemp()
    yield dirname
    # teardown
    shutil.rmtree(dirname)


def test_workspace_init(wspath):
    commands.workspace_init(wspath, {"some": "metadata"})
    try:
        commands.workspace_init(wspath, {"some": "metadata"})
        assert False, "Duplicate workspace init succeeded"
    except errors.CommandError:
        pass


def test_workspace_info(wspath):
    commands.workspace_init(wspath, {"some": "metadata"})
    commands.workspace_info(wspath)

    #    subdir = os.path.join(wspath, 'stuff')
    #    os.mkdir(subdir)
    #    commands.workspace_info(subdir)

    notasubdir = os.path.join(wspath, "nope")
    try:
        commands.workspace_info(notasubdir)
        assert False, "Nonexistent subdir workspace info success"
    except errors.CommandError:
        pass


def test_find_html_docs(wspath):
    filedir = os.path.dirname(__file__)
    docpath = os.path.join(filedir, "..", "docs", "build", "html")
    if os.path.exists(docpath):
        commands.workspace_init(wspath, {}, html_paths=[docpath])
        dmfobj = dmfbase.DMF(wspath)
        filenames = commands.find_html_docs(dmfobj, dmfobj)
        assert len(filenames) > 0



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# This file contains an example application of the RIPE software
# The goal is to identify the reaction kinetics present in a reactor
# where ethylbenzene is converted to styrene, with a number of side products
# 
# Initial training sets provided through random or space-filling sampling 
# result in incorrect model identification
# Error maximization sampling can be used to refine the model

import pyomo.environ as pyo
from idaes.surrogate import ripe
import numpy as np
import random
from . import cracsim

np.random.seed(100)


def main():
    Tr = 750.0
    # Experimental variance is known in this problem,
    # it can be estimated if not provided
    noise = 0.05
    # ndata = 10 in publication example
    ndata = 30
    ns = 9
    # Define temperature bounds
    Tlo = 500
    Tup = 1000
    # Define range of inlet concentrations
    lb = [0,0,0,0,0,1,0,0,1]
    ub = [3]*ns

    gc = .008314

    Temp = np.linspace(Tlo,Tup,ndata)
    # Initialize concentration data

    # Inlet concentrations are fixed in publication example
    # cdata0 = [[.5,0,0,0,0,4.5,0,0,4.5]]*ndata
    cdata0=np.zeros([ndata,ns])
    for i in range(ndata):
        for j in range(ns):
            cdata0[i,j] = random.uniform(lb[j],ub[j])


    # Calculate steady-state concentration values from simulator cracsim.py
    cdata = cracsim.sim(np.hstack((cdata0,np.expand_dims(Temp, axis=1))))

    # In this example, we know the true stoichiometries. Lets define them first for clarity
    t_stoich = [[-1,1,0,0,0,0,1,0,0],[-1,0,1,1,0,0,0,0,0],[0,0,0,-1,2,0,-2,0,0],[0,0,0,0,-1,-2,4,1,0]]
    # Additional considered stoichiometries are defined
    a_stoich = [[-1,0,0,0,0,-16,21,8,0],[-1,0,0,4,0,0,-3,0,0],[0,0,0,-1,0,-4,6,2,0],[-1,0,1,0,2,0,-2,0,0]]
    # Index 0-3 are the true reactions, 4-7 are considered reactions
    stoichs = t_stoich+a_stoich

    # Define kinetic mechanisms, adsorption parameters must be known a-priori
    kco = 35
    kst = 1.5

    # Mechanisms can be defined for each stoichiometry using a list-of-list
    mechs = [[[0,1,3,4,7],eb_dep],[[0],[t_st_prod,cat_st_prod_t1]],[[0,1],cat_ben_prod_t2],[[2,3],meth_prod_t3],[[2,3],ch4_to_co_t4],[[2,3,4,5,6,7],[ma_g,ma_h]]]

    # Experimental variance is known in this case
    sigma = np.multiply(noise**2,cdata)

    results = ripe.ripemodel(cdata,stoich = stoichs,mechanisms=mechs,x0=cdata0,temp=Temp,sigma=sigma,tref=Tr)


def keq(*x):
    a,b,c,d,f,g,h,i,j,T = x
    temp = 0.1 + 300/T
    return pyo.exp(temp)

# These mechanisms are present in the simulation
def cat_st_prod_t1(*x):
    a,b,c,d,f,g,h,i,j,Temp = x
    return (a - (b * h) / keq(*x)) * (1/((1+kst*b)*(1+kco*i)))
def cat_ben_prod_t2(*x):
    # Mechanism for EB > B + C2H4
    a,b,c,d,f,g,h,i,j,Temp = x
    return a / (1+kco*i)
def meth_prod_t3(*x):
    # Mechanism for C2H4+4H2O > 2CO2+6H
    a,b,c,d,f,g,h,i,j,Temp = x
    return d * h
def ch4_to_co_t4(*x):
     a,b,c,d,f,g,h,i,j,Temp = x
    #mechanism for CH4+2H2O > CO2+4H2
     return f * g

# Additional mechanisms are specified for the true stoichiometries and additional stoichs
def eb_dep(*x):
    a,b,c,d,f,g,h,i,j,Temp = x
    return a
def t_st_prod(*x):
    a,b,c,d,f,g,h,i,j,Temp = x
    return (a - (b * h) / keq(*x))
def ma_h(*x):
    a,b,c,d,f,g,h,i,j,Temp = x
    return h
def ma_g(*x):
    a,b,c,d,f,g,h,i,j,Temp = x
    return g


# The following section of code can be un-commented in order to continue sampling data until an accurate model is obtained
# note that this is considerably more computationally expensive
'''
#Append T bounds for ems variables
lb.append(Tlo)
ub.append(Tup)

# Call RIPE ems in order to identify the next best sample point
[new_points, err] = ripe.ems(results,cracsim.sim,lb,ub,ns,x=cdata,x0=cdata0,Temp=Temp,Tref=Tr)#,frac=fracfun)
new_res = cracsim.sim(new_points)
ite = 0
data = cdata
data0 = cdata0
tdata = Temp.tolist()
# print 'maximum allowable tolerances : ', [2*noise*s for s in new_res]
while any(err >  [2*noise*s for s in new_res] ):
#    print 'Which concentration : ', err > [noise*s for s in new_res]
    data = np.vstack((data,new_res))
    data0 = np.vstack((data0,new_points[:-1]))
    tdata.append(new_points[-1])
    results = {}
    ite+=1
    sigma =  np.multiply(noise**2,np.array(data))
    results = ripe.ripemodel(data,stoich = stoichs,mechanisms=mechs,x0=data0,sigma=sigma,temp=tdata,tref=Tr, hide_output=True)
    [new_points, err] = ripe.ems(results,cracsim.sim,lb,ub,10,x=data,x0=data0,temp=tdata,tref=Tr)
    new_res = cracsim.sim(new_points)
#    print 'currently at '+str(len(data))+' data points'
#    print 'proposed new conc : ', new_res
#    print 'maximum allowable tolerances : ', [noise*s for s in new_res]
'''


if __name__ == "__main__":
    main()


##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for the convergence testing module

Author: Carl Laird
"""
import pytest
import os
import os.path
from pyutilib.misc import compare_json_files
import pyomo.environ as pe
from pyomo.common.fileutils import this_file_dir
import idaes.core.util.convergence.convergence_base as cb

# See if ipopt is available and set up solver
ipopt_available = pe.SolverFactory('ipopt').available()
ceval_fixedvar_mutableparam_str = (
        'idaes.core.util.convergence.tests.'
        'conv_eval_classes.ConvEvalFixedVarMutableParam')
ceval_fixedvar_immutableparam_str = (
        'idaes.core.util.convergence.tests.'
        'conv_eval_classes.ConvEvalFixedVarImmutableParam')
ceval_unfixedvar_mutableparam_str = (
        'idaes.core.util.convergence.tests.'
        'conv_eval_classes.ConvEvalUnfixedVarMutableParam')

currdir = this_file_dir()


def test_convergence_evaluation_specification_file_fixedvar_mutableparam():
    ceval_class = cb._class_import(ceval_fixedvar_mutableparam_str)
    ceval = ceval_class()
    import idaes.core.util.convergence.tests.conv_eval_classes as cev
    assert ceval.__class__ == cev.ConvEvalFixedVarMutableParam().__class__

    spec = ceval.get_specification()
    fname = os.path.join(currdir, 'ceval_fixedvar_mutableparam.3.42.json')
    cb.write_sample_file(spec, fname, ceval_fixedvar_mutableparam_str,
                         n_points=3, seed=42)

    baseline_fname = os.path.join(
            currdir,
            'ceval_fixedvar_mutableparam.3.42.baseline.json')
    compare_json_files(baseline_fname=baseline_fname,
                       output_fname=fname,
                       tolerance=1e-8)

    if os.path.exists(fname):
        os.remove(fname)

def test_convergence_evaluation_specification_file_unfixedvar_mutableparam():
    ceval_class = cb._class_import(ceval_unfixedvar_mutableparam_str)
    ceval = ceval_class()
    import idaes.core.util.convergence.tests.conv_eval_classes as cev
    assert ceval.__class__ == cev.ConvEvalUnfixedVarMutableParam().__class__

    spec = ceval.get_specification()
    fname = os.path.join(currdir, 'ceval_unfixedvar_mutableparam.3.42.json')
    cb.write_sample_file(spec, fname,
                         ceval_unfixedvar_mutableparam_str,
                         n_points=3, seed=42)

    baseline_fname = os.path.join(
            currdir,
            'ceval_unfixedvar_mutableparam.3.42.baseline.json')
    compare_json_files(baseline_fname=baseline_fname,
                       output_fname=fname,
                       tolerance=1e-8)

    # expect an exception because var is not fixed
    with pytest.raises(ValueError):
        inputs, samples, global_results = \
            cb.run_convergence_evaluation_from_sample_file(fname)

    if os.path.exists(fname):
        os.remove(fname)


def test_convergence_evaluation_specification_file_fixedvar_immutableparam():
    ceval_class = cb._class_import(ceval_fixedvar_immutableparam_str)
    ceval = ceval_class()
    import idaes.core.util.convergence.tests.conv_eval_classes as cev
    assert ceval.__class__ == cev.ConvEvalFixedVarImmutableParam().__class__

    spec = ceval.get_specification()
    fname = os.path.join(currdir, 'ceval_fixedvar_immutableparam.3.42.json')
    cb.write_sample_file(spec, fname,
                         ceval_fixedvar_immutableparam_str,
                         n_points=3, seed=42)

    baseline_fname = os.path.join(
            currdir,
            'ceval_fixedvar_immutableparam.3.42.baseline.json')
    compare_json_files(baseline_fname=baseline_fname,
                       output_fname=fname,
                       tolerance=1e-8)

    # expect an exception because param is not mutable
    with pytest.raises(ValueError):
        inputs, samples, global_results = \
            cb.run_convergence_evaluation_from_sample_file(fname)

    if os.path.exists(fname):
        os.remove(fname)


@pytest.mark.skipif(ipopt_available == False,
                    reason="Ipopt solver not available")
def test_convergence_evaluation_fixedvar_mutableparam():
    ceval_class = cb._class_import(ceval_fixedvar_mutableparam_str)
    ceval = ceval_class()
    import idaes.core.util.convergence.tests.conv_eval_classes as cev
    assert ceval.__class__ == cev.ConvEvalFixedVarMutableParam().__class__

    spec = ceval.get_specification()
    fname = os.path.join(currdir, 'ceval_fixedvar_mutableparam.3.43.json')
    cb.write_sample_file(spec, fname, ceval_fixedvar_mutableparam_str,
                         n_points=3, seed=43)

    inputs, samples, global_results = \
        cb.run_convergence_evaluation_from_sample_file(fname)

    # put the results into a json file for comparison
    # jsondict = dict(inputs=inputs, samples=samples,
    #                 global_results=global_results)
    # results_fname = os.path.join(
    #        currdir, 'ceval_fixedvar_mutableparam.3.43.results.json')
    # with open(results_fname, 'w') as fd:
    #     json.dump(jsondict, fd, indent=3)

    # compare results
    assert global_results[0]['name'] == 'Sample-1'
    assert global_results[0]['solved']
    assert global_results[0]['iters'] == 14

    assert global_results[1]['name'] == 'Sample-2'
    assert global_results[1]['solved']
    assert global_results[1]['iters'] == 15

    assert global_results[2]['name'] == 'Sample-3'
    assert global_results[2]['solved']
    assert global_results[2]['iters'] == 12

    if os.path.exists(fname):
        os.remove(fname)
    # if os.path.exists(results_fname):
    #     os.remove(results_fname)


if __name__ == '__main__':
    # test_convergence_evaluation_specification_file_fixedvar_mutableparam()
    # test_convergence_evaluation_specification_file_unfixedvar_mutableparam()
    # test_convergence_evaluation_specification_file_fixedvar_immutableparam()
    test_convergence_evaluation_fixedvar_mutableparam()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for saponification property package example.
Authors: Andrew Lee
"""
import logging
import pytest
from pyomo.environ import (ConcreteModel,
                           Constraint,
                           Param,
                           value,
                           Var)
from idaes.core import MaterialFlowBasis

from idaes.property_models.examples.saponification_reactions import (
    SaponificationReactionParameterBlock, ReactionBlock)
from idaes.property_models.examples.saponification_thermo import (
    SaponificationParameterBlock)

from idaes.core.util.testing import get_default_solver


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


class TestParamBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.pparams = SaponificationParameterBlock()
        model.rparams = SaponificationReactionParameterBlock(
                default={"property_package": model.pparams})

        return model

    def test_config(self, model):
        assert len(model.rparams.config) == 2

    def test_build(self, model):
        assert model.rparams.reaction_block_class is ReactionBlock

        assert len(model.rparams.phase_list) == 1
        for i in model.rparams.phase_list:
            assert i == "Liq"

        assert len(model.rparams.component_list) == 5
        for i in model.rparams.component_list:
            assert i in ['H2O',
                         'NaOH',
                         'EthylAcetate',
                         'SodiumAcetate',
                         'Ethanol']

        assert len(model.rparams.rate_reaction_idx) == 1
        for i in model.rparams.rate_reaction_idx:
            assert i == "R1"

        assert len(model.rparams.rate_reaction_stoichiometry) == 5
        for i in model.rparams.rate_reaction_stoichiometry:
            assert i in [("R1", "Liq", "NaOH"),
                         ("R1", "Liq", "EthylAcetate"),
                         ("R1", "Liq", "SodiumAcetate"),
                         ("R1", "Liq", "Ethanol"),
                         ("R1", "Liq", "H2O")]

        assert isinstance(model.rparams.arrhenius, Param)
        assert value(model.rparams.arrhenius) == 3.132e6

        assert isinstance(model.rparams.energy_activation, Param)
        assert value(model.rparams.energy_activation) == 43000

        assert isinstance(model.rparams.dh_rxn, Param)
        assert len(model.rparams.dh_rxn) == 1
        for i in model.rparams.dh_rxn:
            assert value(model.rparams.dh_rxn[i]) == -49000

        assert isinstance(model.rparams.gas_const, Param)
        assert value(model.rparams.gas_const) == 8.314


class TestReactionBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.pparams = SaponificationParameterBlock()
        model.rparams = SaponificationReactionParameterBlock(
                default={"property_package": model.pparams})

        model.props = model.pparams.state_block_class(
                [1],
                default={"parameters": model.pparams})

        model.rxns = model.rparams.reaction_block_class(
                [1],
                default={"parameters": model.rparams,
                         "state_block": model.props})

        return model

    def test_build(self, model):
        assert model.rxns[1].conc_mol_comp_ref is model.props[1].conc_mol_comp
        assert model.rxns[1].temperature_ref is model.props[1].temperature
        assert model.rxns[1].dh_rxn is model.rparams.dh_rxn

    def test_rate_constant(self, model):
        assert isinstance(model.rxns[1].k_rxn, Var)
        assert isinstance(model.rxns[1].arrhenius_eqn, Constraint)

    def test_rxn_rate(self, model):
        assert isinstance(model.rxns[1].reaction_rate, Var)
        assert isinstance(model.rxns[1].rate_expression, Constraint)

    def test_get_reaction_rate_basis(self, model):
        assert model.rxns[1].get_reaction_rate_basis() == \
            MaterialFlowBasis.molar

    def test_model_check(self, model):
        assert model.rxns[1].model_check() is None

    def test_initialize(self, model):
        assert model.rxns.initialize(outlvl=1) is None



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import pyomo.environ as pyo
from idaes.surrogate import ripe
import numpy as np
import random
from . import isotsim

np.random.seed(20)


def main():

    #ndata = 100
    noise = 0.1
    ns = 5
    lb_conc = [0,0,0,0,0]
    ub_conc = [10,10,0,0,0]


    # Initialize concentration arrays

    # initial concentrations - only 2 data points at bounds
    cdata0 = [[1,1,0,0,0],[10,10,0,0,0]]
    cdata = isotsim.sim(cdata0)
    nd = len(cdata0)

    # Considered reaction stoichiometries
    stoich = [[-1,-1,1,0,0] ,[0,-1,-1,1,0],[-1,0,0,-1,1],[-1,-2,0,1,0] ,[-2,-2,0,0,1],[-1,-1,-1,0,1],[-2,-1,1,-1,1]]

    # IRIPE internal mass action kinetics are specified
    mechs = [['all','massact']]

    # Use expected variance - estimated from data if not provided
    sigma = np.multiply(noise**2,np.array(cdata))

    # Call to RIPE
    results = ripe.ripemodel(cdata,stoich = stoich,mechanisms=mechs,x0=cdata0,hide_output=False,sigma=sigma,deltaterm=0,expand_output=True)

    # Adaptive experimental design using error maximization sampling
    [new_points, err] = ripe.ems(results,isotsim.sim,lb_conc,ub_conc,5,x=cdata,x0=cdata0)

    # Implement EMS as described in the RIPE publication
    new_res = isotsim.sim(new_points)[0]
    ite = 0
    # print 'maximum allowable tolerances : ', [noise*s for s in new_res]
    while any(err >  [2*noise*s for s in new_res] ):
    #    print 'Which concentrations violate error (True=violation) : ', err > [noise*s for s in new_res]
        results = {}
        ite+=1
        # Data updated explicitly so RBFopt subroutines produce consistent results
        new_cdata0 = np.zeros([nd+ite,ns])
        new_cdata  = np.zeros([nd+ite,ns])
        new_cdata0[:-1][:] = cdata0[:][:]
        new_cdata[:-1][:] = cdata[:][:]
        new_cdata0[-1][:] = new_points[:]
        res = isotsim.sim(new_points)[0]
        for j in range(len(res)):
            new_cdata[-1][j] = res[j]

        #Update weight parameters
        sigma =  np.multiply(noise**2,np.array(new_cdata))

        # Build updated RIPE model
        results = ripe.ripemodel(new_cdata,stoich = stoich,mechanisms=mechs,x0=new_cdata0,sigma=sigma,expand_output=True)

        # Another call to EMS
        [new_points, err] = ripe.ems(results,isotsim.sim,lb_conc,ub_conc,5,x=cdata,x0=cdata0)

        # Update results
        new_res = isotsim.sim(new_points)[0]
        cdata0 = new_cdata0
        cdata = new_cdata

    # Final call to RIPE to get concise output
    results = ripe.ripemodel(cdata,stoich = stoich,mechanisms=mechs,x0=cdata0,sigma=sigma,expand_output=False)
    #print results


if __name__ == "__main__":
    main()






##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for idaes.dmf.propdb.types
"""
# stdlib
import json
import logging
import os
import sys
from io import StringIO

# third-party
import pytest

# local
from idaes.dmf.propdata import PropertyTable, PropertyData, PropertyMetadata
from idaes.dmf.util import TempDir

# for testing
from .util import init_logging

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)


# keep minimal example first for test_Property{Data,Metadata}Parsing
good_data_csv = [
    # minimal
    "Num,State (units 1),Absolute Error,Prop (units 2),Relative Error\n"
    "1,1.0,0,100.0,0.1",
    # two values, "T" and "Density"
    "Data No.,T (K),Absolute Error,Density (g/cm3),Absolute Error\n"
    "1,303.15,0,0.2,0\n"
    "2,304.15,0,0.3,0\n",
    # same but with missing values
    "Data No.,T (K),Absolute Error,Density (g/cm3),Absolute Error\n"
    "1,303.15,,0.2,0\n"
    "2,304.15,0,,0\n",
]

bad_data_csv = ["", "blah, blah", "blah\nblah", "N,X\n1,1,0"]  # missing error in hdr

good_metadata_text = [
    # minimal
    'source , Mr. Author,"A Title",Other Info',
    # Full set of fields
    'Source,Han, J., Jin, J., Eimer, D.A., Melaaen, M.C.,"Density of '
    "Water(1) + Monoethanolamine(2) + CO2(3) from (298.15 to 413.15) "
    "K and Surface Tension of Water(1) + Monethanolamine(2) from "
    '(303.15 to 333.15)K", J. Chem. Eng. Data, 2012, Vol. 57, pg. '
    '1095-1103"\n'
    'Retrieval,"J. Morgan, date unknown"\n'
    "Notes,r is MEA weight fraction in aqueous soln. (CO2-free basis)",
]

good_table_json = [
    {
        "meta": [
            {
                "datatype": "MEA",
                "info": "J. Chem. Eng. Data, 2009, Vol 54, pg. 3096-30100",
                "notes": "r is MEA weight fraction in aqueous soln.",
                "authors": "Amundsen, T.G., Lars, E.O., Eimer, D.A.",
                "title": "Density and Viscosity of Monoethanolamine + .etc.",
                "date": "2009",
            }
        ],
        "data": [
            {
                "name": "Viscosity Value",
                "units": "mPa-s",
                "values": [2.6, 6.2],
                "error_type": "absolute",
                "errors": [0.06, 0.004],
                "type": "property",
            },
            {"name": "r", "units": "", "values": [0.2, 1000], "type": "state"},
        ],
    }
]

dummy_data = [{"name": "dummy", "type": "state", "units": "none", "values": [1, 2, 3]}]


def test_property_table_empty():
    with pytest.raises(ValueError):
        PropertyTable(data=[], metadata=[])


def test_property_table():
    tbl = PropertyTable(data=dummy_data, metadata=[])
    assert tbl.data is not None
    assert tbl.metadata is not None


def test_property_table_objinit():
    d = PropertyData(dummy_data)
    m = PropertyMetadata({"bar": "y"})
    tbl = PropertyTable(data=d, metadata=m)
    assert tbl.data.as_list()[0]["units"] == "none"
    assert tbl.metadata[0].as_dict()["bar"] == "y"


def test_property_table_json():
    tbl = PropertyTable(data=dummy_data, metadata={"baz": "2"})
    s = tbl.dumps()
    parsed = json.loads(s)
    assert parsed["data"] == tbl.data.as_list()
    with TempDir() as d:
        fp = open(os.path.join(d, "tbl.json"), "w")
        tbl.dump(fp)
        fp.close()
        fp = open(os.path.join(d, "tbl.json"), "r")
        parsed = json.load(fp)
    assert parsed["data"] == tbl.data.as_list()


def test_property_table_from_json():
    with TempDir() as d:
        filename = os.path.join(d, "prop.json")
        for gj in good_table_json:
            fp = open(filename, "w")
            json.dump(gj, fp)
            fp.close()
            fp = open(filename, "r")
            tbl = PropertyTable.load(fp)
            assert tbl.data.num_columns == 2
        os.unlink(filename)


def test_property_data_good():
    inputs = map(StringIO, good_data_csv)
    states, total = [1], [2]
    for inpfile, state_col, ncol in zip(inputs, states, total):
        obj = PropertyData.from_csv(inpfile, state_col)
        assert obj.num_columns == ncol


def test_property_data_column_type():
    csv_input = good_data_csv[0]
    obj = PropertyData.from_csv(StringIO(csv_input), 1)
    csv_header = csv_input.split("\n")[0].split(",")
    for i, col in enumerate((csv_header[1], csv_header[3])):
        units_pos = col.find("(")
        if units_pos >= 0:
            col = col[:units_pos].strip()
        obj.get_column(col)
        index = obj.get_column_index(col)
        if i == 0:
            assert obj.is_state_column(index)
        else:
            assert obj.is_property_column(index)


def test_property_data_column_nonexistent():
    csv_input = good_data_csv[0]
    obj = PropertyData.from_csv(StringIO(csv_input), 1)
    try:
        obj.get_column("this is not a column")
        assert False, "Expected failure for bogus column"
    except KeyError:
        pass


def test_property_data_bad():
    inputs = map(StringIO, bad_data_csv)
    for inpfile in inputs:
        with pytest.raises(Exception):
            PropertyData.from_csv(inpfile, 0)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Basic IDAES 1D Heat Exchanger Model.

1D Single pass shell and tube HX model with 0D wall conduction model
"""
# Import Python libraries
import math
from enum import Enum

# Import Pyomo libraries
from pyomo.environ import (SolverFactory, Var, Param, Constraint,
                           value, TerminationCondition)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (ControlVolume1DBlock,
                        UnitModelBlockData,
                        declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        FlowDirection,
                        useDefault)
from idaes.unit_models.heat_exchanger import HeatExchangerFlowPattern
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition


__author__ = "Jaffer Ghouse"

# Set up logger
_log = getIdaesLogger(__name__)


class WallConductionType(Enum):
    zero_dimensional = 0
    one_dimensional = 1
    two_dimensional = 2


@declare_process_block_class("HeatExchanger1D")
class HeatExchanger1DData(UnitModelBlockData):
    """Standard Heat Exchanger 1D Unit Model Class."""

    CONFIG = UnitModelBlockData.CONFIG()
    # Template for config arguments for shell and tube side
    _SideTemplate = ConfigBlock()
    _SideTemplate.declare("dynamic", ConfigValue(
        default=useDefault,
        domain=In([useDefault, True, False]),
        description="Dynamic model flag",
        doc="""Indicates whether this model will be dynamic or not,
**default** = useDefault.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - set as a dynamic model,
**False** - set as a steady-state model.}"""))
    _SideTemplate.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Holdup construction flag",
        doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**True** - construct holdup terms,
**False** - do not construct holdup terms}"""))
    _SideTemplate.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    _SideTemplate.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    _SideTemplate.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    _SideTemplate.declare("has_pressure_change", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Pressure change term construction flag",
        doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}"""))
    _SideTemplate.declare("has_phase_equilibrium", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Phase equilibrium term construction flag",
        doc="""Argument to enable phase equilibrium on the shell side.
- True - include phase equilibrium term
- False - do not include phase equilibrium term"""))
    _SideTemplate.declare("property_package", ConfigValue(
        default=None,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a ParameterBlock object"""))
    _SideTemplate.declare("property_package_args", ConfigValue(
        default={},
        description="Arguments for constructing shell property package",
        doc="""A dict of arguments to be passed to the PropertyBlockData
and used when constructing these
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a dict (see property package for documentation)"""))
    # TODO : We should probably think about adding a consistency check for the
    # TODO : discretisation methods as well.
    _SideTemplate.declare("transformation_method", ConfigValue(
        default=useDefault,
        description="Discretization method to use for DAE transformation",
        doc="""Discretization method to use for DAE transformation. See Pyomo
documentation for supported transformations."""))
    _SideTemplate.declare("transformation_scheme", ConfigValue(
        default=useDefault,
        description="Discretization scheme to use for DAE transformation",
        doc="""Discretization scheme to use when transformating domain. See
Pyomo documentation for supported schemes."""))

    # Create individual config blocks for shell and tube side
    CONFIG.declare("shell_side",
                   _SideTemplate(doc="shell side config arguments"))
    CONFIG.declare("tube_side",
                   _SideTemplate(doc="tube side config arguments"))

    # Common config args for both sides
    CONFIG.declare("finite_elements", ConfigValue(
        default=20,
        domain=int,
        description="Number of finite elements length domain",
        doc="""Number of finite elements to use when discretizing length
domain (default=20)"""))
    CONFIG.declare("collocation_points", ConfigValue(
        default=5,
        domain=int,
        description="Number of collocation points per finite element",
        doc="""Number of collocation points to use per finite element when
discretizing length domain (default=3)"""))
    CONFIG.declare("flow_type", ConfigValue(
        default=HeatExchangerFlowPattern.cocurrent,
        domain=In(HeatExchangerFlowPattern),
        description="Flow configuration of heat exchanger",
        doc="""Flow configuration of heat exchanger
- HeatExchangerFlowPattern.cocurrent: shell and tube flows from 0 to 1
(default)
- HeatExchangerFlowPattern.countercurrent: shell side flows from 0 to 1
tube side flows from 1 to 0"""))
    CONFIG.declare("has_wall_conduction", ConfigValue(
        default=WallConductionType.zero_dimensional,
        domain=In(WallConductionType),
        description="Conduction model for tube wall",
        doc="""Argument to enable type of wall heat conduction model.
- WallConductionType.zero_dimensional - 0D wall model (default),
- WallConductionType.one_dimensional - 1D wall model along the thickness of the
tube,
- WallConductionType.two_dimensional - 2D wall model along the lenghth and
thickness of the tube"""))

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(HeatExchanger1DData, self).build()

        # Set flow directions for the control volume blocks and specify
        # dicretisation if not specified.
        if self.config.flow_type == HeatExchangerFlowPattern.cocurrent:
            set_direction_shell = FlowDirection.forward
            set_direction_tube = FlowDirection.forward
            if (self.config.shell_side.transformation_method !=
                self.config.tube_side.transformation_method) or \
                (self.config.shell_side.transformation_scheme !=
                    self.config.tube_side.transformation_scheme):
                raise ConfigurationError(
                    "HeatExchanger1D only supports similar transformation "
                    "schemes on the shell side and tube side domains for "
                    "both cocurrent and countercurrent flow patterns.")
            if self.config.shell_side.transformation_method is useDefault:
                _log.warning("Discretization method was "
                             "not specified for the shell side of the "
                             "co-current heat exchanger. "
                             "Defaulting to finite "
                             "difference method on the shell side.")
                self.config.shell_side.transformation_method = \
                    "dae.finite_difference"
            if self.config.tube_side.transformation_method is useDefault:
                _log.warning("Discretization method was "
                             "not specified for the tube side of the "
                             "co-current heat exchanger. "
                             "Defaulting to finite "
                             "difference method on the tube side.")
                self.config.tube_side.transformation_method = \
                    "dae.finite_difference"
            if self.config.shell_side.transformation_scheme is useDefault:
                _log.warning("Discretization scheme was "
                             "not specified for the shell side of the "
                             "co-current heat exchanger. "
                             "Defaulting to backward finite "
                             "difference on the shell side.")
                self.config.shell_side.transformation_scheme = "BACKWARD"
            if self.config.tube_side.transformation_scheme is useDefault:
                _log.warning("Discretization scheme was "
                             "not specified for the tube side of the "
                             "co-current heat exchanger. "
                             "Defaulting to backward finite "
                             "difference on the tube side.")
                self.config.tube_side.transformation_scheme = "BACKWARD"
        elif self.config.flow_type == HeatExchangerFlowPattern.countercurrent:
            set_direction_shell = FlowDirection.forward
            set_direction_tube = FlowDirection.backward
            if self.config.shell_side.transformation_method is useDefault:
                _log.warning("Discretization method was "
                             "not specified for the shell side of the "
                             "counter-current heat exchanger. "
                             "Defaulting to finite "
                             "difference method on the shell side.")
                self.config.shell_side.transformation_method = \
                    "dae.finite_difference"
            if self.config.tube_side.transformation_method is useDefault:
                _log.warning("Discretization method was "
                             "not specified for the tube side of the "
                             "counter-current heat exchanger. "
                             "Defaulting to finite "
                             "difference method on the tube side.")
                self.config.tube_side.transformation_method = \
                    "dae.finite_difference"
            if self.config.shell_side.transformation_scheme is useDefault:
                _log.warning("Discretization scheme was "
                             "not specified for the shell side of the "
                             "counter-current heat exchanger. "
                             "Defaulting to backward finite "
                             "difference on the shell side.")
                self.config.shell_side.transformation_scheme = "BACKWARD"
            if self.config.tube_side.transformation_scheme is useDefault:
                _log.warning("Discretization scheme was "
                             "not specified for the tube side of the "
                             "counter-current heat exchanger. "
                             "Defaulting to forward finite "
                             "difference on the tube side.")
                self.config.tube_side.transformation_scheme = "BACKWARD"
        else:
            raise ConfigurationError(
                    "{} HeatExchanger1D only supports cocurrent and "
                    "countercurrent flow patterns, but flow_type configuration"
                    " argument was set to {}."
                    .format(self.name,
                            self.config.flow_type))

        # Control volume 1D for shell
        self.shell = ControlVolume1DBlock(default={
            "dynamic": self.config.shell_side.dynamic,
            "has_holdup": self.config.shell_side.has_holdup,
            "property_package": self.config.shell_side.property_package,
            "property_package_args":
                self.config.shell_side.property_package_args,
            "transformation_method": self.config.shell_side.transformation_method,
            "transformation_scheme": self.config.shell_side.transformation_scheme,
            "finite_elements": self.config.finite_elements,
            "collocation_points": self.config.collocation_points})

        self.tube = ControlVolume1DBlock(default={
            "dynamic": self.config.tube_side.dynamic,
            "has_holdup": self.config.tube_side.has_holdup,
            "property_package": self.config.tube_side.property_package,
            "property_package_args":
                self.config.tube_side.property_package_args,
            "transformation_method": self.config.tube_side.transformation_method,
            "transformation_scheme": self.config.tube_side.transformation_scheme,
            "finite_elements": self.config.finite_elements,
            "collocation_points": self.config.collocation_points})

        self.shell.add_geometry(flow_direction=set_direction_shell)
        self.tube.add_geometry(flow_direction=set_direction_tube)

        self.shell.add_state_blocks(
            information_flow=set_direction_shell,
            has_phase_equilibrium=self.config.shell_side.has_phase_equilibrium)
        self.tube.add_state_blocks(
            information_flow=set_direction_tube,
            has_phase_equilibrium=self.config.tube_side.has_phase_equilibrium)

        # Populate shell
        self.shell.add_material_balances(
            balance_type=self.config.shell_side.material_balance_type,
            has_phase_equilibrium=self.config.shell_side.has_phase_equilibrium)

        self.shell.add_energy_balances(
            balance_type=self.config.shell_side.energy_balance_type,
            has_heat_transfer=True)

        self.shell.add_momentum_balances(
            balance_type=self.config.shell_side.momentum_balance_type,
            has_pressure_change=self.config.shell_side.has_pressure_change)

        self.shell.apply_transformation()

        # Populate tube
        self.tube.add_material_balances(
            balance_type=self.config.tube_side.material_balance_type,
            has_phase_equilibrium=self.config.tube_side.has_phase_equilibrium)

        self.tube.add_energy_balances(
            balance_type=self.config.tube_side.energy_balance_type,
            has_heat_transfer=True)

        self.tube.add_momentum_balances(
            balance_type=self.config.tube_side.momentum_balance_type,
            has_pressure_change=self.config.tube_side.has_pressure_change)

        self.tube.apply_transformation()

        # Add Ports for shell side
        self.add_inlet_port(name="shell_inlet", block=self.shell)
        self.add_outlet_port(name="shell_outlet", block=self.shell)

        # Add Ports for tube side
        self.add_inlet_port(name="tube_inlet", block=self.tube)
        self.add_outlet_port(name="tube_outlet", block=self.tube)

        # Add reference to control volume geometry
        add_object_reference(self,
                             "shell_area",
                             self.shell.area)
        add_object_reference(self,
                             "shell_length",
                             self.shell.length)
        add_object_reference(self,
                             "tube_area",
                             self.tube.area)
        add_object_reference(self,
                             "tube_length",
                             self.tube.length)

        self._make_performance()

    def _make_performance(self):
        """
        Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        # Unit model parameters
        self.pi = Param(initialize=math.pi, doc="pi")

        # Unit model variables
        # HX dimensions
        self.d_shell = Var(initialize=1, doc="Diameter of shell")
        self.d_tube_outer = Var(initialize=0.011, doc="Outer diameter of tube")
        self.d_tube_inner = Var(initialize=0.010, doc="Inner diameter of tube")
        self.N_tubes = Var(initialize=1, doc="Number of tubes")

        # Note: In addition to the above variables, "shell_length" and
        # "tube_length" need to be fixed at the flowsheet level

        # Performance variables
        self.shell_heat_transfer_coefficient = Var(
                self.flowsheet().config.time,
                self.shell.length_domain,
                initialize=50,
                doc="Heat transfer coefficient")
        self.tube_heat_transfer_coefficient = Var(self.flowsheet().config.time,
                                                  self.tube.length_domain,
                                                  initialize=50,
                                                  doc="Heat transfer "
                                                  "coefficient")

        # Wall 0D model (Q_shell = Q_tube*N_tubes)
        if (self.config.has_wall_conduction ==
                WallConductionType.zero_dimensional):
            self.temperature_wall = Var(self.flowsheet().config.time,
                                        self.tube.length_domain,
                                        initialize=298.15)

            # Performance equations
            # Energy transfer between shell and tube wall

            @self.Constraint(self.flowsheet().config.time,
                             self.shell.length_domain,
                             doc="Heat transfer between shell and tube")
            def shell_heat_transfer_eq(self, t, x):
                return self.shell.heat[t, x] == - self.N_tubes *\
                    (self.shell_heat_transfer_coefficient[t, x] *
                     self.pi * self.d_tube_outer *
                     (self.shell.properties[t, x].temperature -
                      self.temperature_wall[t, x]))

            # Energy transfer between tube wall and tube
            @self.Constraint(self.flowsheet().config.time,
                             self.tube.length_domain,
                             doc="Convective heat transfer")
            def tube_heat_transfer_eq(self, t, x):
                return self.tube.heat[t, x] == \
                    self.tube_heat_transfer_coefficient[t, x] *\
                    self.pi * self.d_tube_inner * \
                    (self.temperature_wall[t, x] -
                     self.tube.properties[t, x].temperature)

            # Wall 0D model
            @self.Constraint(self.flowsheet().config.time,
                             self.shell.length_domain,
                             doc="wall 0D model")
            def wall_0D_model(self, t, x):
                return self.tube.heat[t, x] == -(self.shell.heat[t, x] /
                                                 self.N_tubes)

        else:
            raise NotImplementedError(
                    "{} HeatExchanger1D has not yet implemented support for "
                    "wall conduction models.")

        # Define tube area in terms of tube diameter
        self.area_calc_tube = Constraint(expr=4 * self.tube_area == self.pi *
                                         self.d_tube_inner**2)

        # Define shell area in terms of shell and tube diameter
        self.area_calc_shell = Constraint(expr=4 * self.shell_area ==
                                          self.pi * (self.d_shell**2 -
                                                     self.N_tubes *
                                                     self.d_tube_outer**2))

    def initialize(blk, shell_state_args=None, tube_state_args=None, outlvl=5,
                   solver='ipopt', optarg={'tol': 1e-6}):
        """
        Initialization routine for the unit (default solver ipopt).

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            optarg : solver options dictionary object (default={'tol': 1e-6})
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')

        Returns:
            None
        """
        init_log = getInitLogger(blk.name, outlvl)

        opt = SolverFactory(solver)
        opt.options = optarg

        # ---------------------------------------------------------------------
        # Initialize shell block
        flags_shell = blk.shell.initialize(outlvl=outlvl + 1,
                                           optarg=optarg,
                                           solver=solver,
                                           state_args=shell_state_args)

        flags_tube = blk.tube.initialize(outlvl=outlvl + 1,
                                         optarg=optarg,
                                         solver=solver,
                                         state_args=tube_state_args)

        init_log.log(4, 'Initialization Step 1 Complete.')

        # ---------------------------------------------------------------------
        # Solve unit
        # Wall 0D
        if blk.config.has_wall_conduction == \
                WallConductionType.zero_dimensional:
            for t in blk.flowsheet().config.time:
                for z in blk.shell.length_domain:
                    blk.temperature_wall[t, z].fix(value(
                        0.5 * (blk.shell.properties[t, 0].temperature +
                               blk.tube.properties[t, 0].temperature)))

            blk.tube.deactivate()
            blk.tube_heat_transfer_eq.deactivate()
            blk.wall_0D_model.deactivate()

            results = opt.solve(blk, tee=init_tee(init_log))
            init_log.log(4, "Initialization Step 2 {}.".format(condition(results)))

            blk.tube.activate()
            blk.tube_heat_transfer_eq.activate()

            results = opt.solve(blk, tee=init_tee(init_log))
            init_log.log(4, "Initialization Step 3 {}.".format(condition(results)))

            blk.wall_0D_model.activate()
            blk.temperature_wall.unfix()

            results = opt.solve(blk, tee=init_tee(init_log))
            init_log.log(4, "Initialization Step 4 {}.".format(condition(results)))


        blk.shell.release_state(flags_shell)
        blk.tube.release_state(flags_tube)

        init_log.log(5, "Initialization Complete.")


    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Shell Area"] = self.shell.area
        var_dict["Shell Diameter"] = self.d_shell
        var_dict["Shell Length"] = self.shell.length
        var_dict["Tube Area"] = self.tube.area
        var_dict["Tube Outer Diameter"] = self.d_tube_outer
        var_dict["Tube Inner Diameter"] = self.d_tube_inner
        var_dict["Tube Length"] = self.tube.length
        var_dict["Number of Tubes"] = self.N_tubes

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
                {"Shell Inlet": self.shell_inlet,
                 "Shell Outlet": self.shell_outlet,
                 "Tube Inlet": self.tube_inlet,
                 "Tube Outlet": self.tube_outlet},
                time_point=time_point)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Ideal property package with VLE calucations. Correlations to compute
Cp_comp, h_comp and vapor pressure are obtained from "The properties of gases
and liquids by Robert C. Reid" and "Perry's Chemical Engineers Handbook by
Robert H. Perry". SI units.
"""

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import Constraint, Expression, log, NonNegativeReals, \
    value, Var, Set, Param, sqrt, log10
from pyomo.opt import SolverFactory, TerminationCondition
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.common.config import ConfigValue, In

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        MaterialFlowBasis,
                        PhysicalParameterBlock,
                        StateBlockData,
                        StateBlock,
                        MaterialBalanceType,
                        EnergyBalanceType)
from idaes.core.util.initialization import (fix_state_vars,
                                            revert_state_vars,
                                            solve_indexed_blocks)
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import BurntToast, ConfigurationError
from idaes.core.util.model_statistics import degrees_of_freedom

# Some more inforation about this module
__author__ = "Jaffer Ghouse"
__version__ = "0.0.2"


# Set up logger
_log = logging.getLogger(__name__)


class IdealParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class
    Contains parameters and indexing sets associated with properties for
    BTX system.
    """
    # Config block for the _IdealStateBlock
    CONFIG = PhysicalParameterBlock.CONFIG()

    CONFIG.declare("valid_phase", ConfigValue(
        default=('Vap', 'Liq'),
        domain=In(['Liq', 'Vap', ('Vap', 'Liq'), ('Liq', 'Vap')]),
        description="Flag indicating the valid phase",
        doc="""Flag indicating the valid phase for a given set of
conditions, and thus corresponding constraints  should be included,
**default** - ('Vap', 'Liq').
**Valid values:** {
**'Liq'** - Liquid only,
**'Vap'** - Vapor only,
**('Vap', 'Liq')** - Vapor-liquid equilibrium,
**('Liq', 'Vap')** - Vapor-liquid equilibrium,}"""))

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(IdealParameterData, self).build()

        self.state_block_class = IdealStateBlock

        # List of valid phases in property package
        if self.config.valid_phase == ('Liq', 'Vap') or \
                self.config.valid_phase == ('Vap', 'Liq'):
            self.phase_list = Set(initialize=['Liq', 'Vap'],
                                  ordered=True)
        elif self.config.valid_phase == 'Liq':
            self.phase_list = Set(initialize=['Liq'])
        else:
            self.phase_list = Set(initialize=['Vap'])

    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.add_properties(
            {'flow_mol': {'method': None, 'units': 'mol/s'},
             'mole_frac_comp': {'method': None, 'units': 'none'},
             'temperature': {'method': None, 'units': 'K'},
             'pressure': {'method': None, 'units': 'Pa'},
             'flow_mol_phase': {'method': None, 'units': 'mol/s'},
             'dens_mol_phase': {'method': '_dens_mol_phase',
                                'units': 'mol/m^3'},
             'pressure_sat': {'method': '_pressure_sat', 'units': 'Pa'},
             'mole_frac_phase_comp': {'method': '_mole_frac_phase',
                                      'units': 'no unit'},
             'energy_internal_mol_phase_comp': {
                     'method': '_energy_internal_mol_phase_comp',
                     'units': 'J/mol'},
             'energy_internal_mol_phase': {
                     'method': '_enenrgy_internal_mol_phase',
                     'units': 'J/mol'},
             'enth_mol_phase_comp': {'method': '_enth_mol_phase_comp',
                                     'units': 'J/mol'},
             'enth_mol_phase': {'method': '_enth_mol_phase',
                                'units': 'J/mol'},
             'entr_mol_phase_comp': {'method': '_entr_mol_phase_comp',
                                     'units': 'J/mol'},
             'entr_mol_phase': {'method': '_entr_mol_phase',
                                'units': 'J/mol'},
             'temperature_bubble': {'method': '_temperature_bubble',
                                    'units': 'K'},
             'temperature_dew': {'method': '_temperature_dew',
                                 'units': 'K'},
             'pressure_bubble': {'method': '_pressure_bubble',
                                 'units': 'Pa'},
             'pressure_dew': {'method': '_pressure_dew',
                              'units': 'Pa'},
             'fug_vap': {'method': '_fug_vap', 'units': 'Pa'},
             'fug_liq': {'method': '_fug_liq', 'units': 'Pa'},
             'dh_vap': {'method': '_dh_vap', 'units': 'J/mol'},
             'ds_vap': {'method': '_ds_vap', 'units': 'J/mol.K'}})

        obj.add_default_units({'time': 's',
                               'length': 'm',
                               'mass': 'g',
                               'amount': 'mol',
                               'temperature': 'K',
                               'energy': 'J',
                               'holdup': 'mol'})


class _IdealStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(blk, state_args=None, state_vars_fixed=False,
                   hold_state=False, outlvl=1,
                   solver='ipopt', optarg={'tol': 1e-8}):
        """
        Initialization routine for property package.
        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provied at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:

                         flow_mol : value at which to initialize component
                                    flows
                         pressure : value at which to initialize pressure
                         temperature : value at which to initialize temperature
                         mole_frac_comp: value at which to initialize the
                                     component mixture mole fraction
            outlvl : sets output level of initialization routine
                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = include solver output infomation (tee=True)
            optarg : solver options dictionary object (default=None)
            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed by the
                                       control volume 1D. Control volume 0D
                                       does not fix the state vars, so will
                                       be False if this state block is used
                                       with 0D blocks.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """

        _log.info('Starting {} initialization'.format(blk.name))

        # Deactivate the constraints specific for outlet block i.e.
        # when defined state is False
        for k in blk.keys():
            if blk[k].config.defined_state is False:
                blk[k].sum_mole_frac_out.deactivate()

        # Fix state variables if not already fixed
        if state_vars_fixed is False:
            flags = fix_state_vars(blk, state_args)

        else:
            # Check when the state vars are fixed already result in dof 0
            for k in blk.keys():
                if degrees_of_freedom(blk[k]) != 0:
                    raise Exception("State vars fixed but degrees of freedom "
                                    "for state block is not zero during "
                                    "initialization.")
        # Set solver options
        if outlvl > 1:
            stee = True
        else:
            stee = False

        if optarg is None:
            sopt = {'tol': 1e-8}
        else:
            sopt = optarg

        opt = SolverFactory('ipopt')
        opt.options = sopt

        # ---------------------------------------------------------------------
        # If present, initialize bubble and dew point calculations
        for k in blk.keys():
            if hasattr(blk[k], "eq_temperature_bubble"):
                calculate_variable_from_constraint(
                        blk[k].temperature_bubble,
                        blk[k].eq_temperature_bubble)

            if hasattr(blk[k], "eq_temperature_dew"):
                calculate_variable_from_constraint(blk[k].temperature_dew,
                                                   blk[k].eq_temperature_dew)

            if hasattr(blk[k], "eq_pressure_bubble"):
                calculate_variable_from_constraint(blk[k].pressure_bubble,
                                                   blk[k].eq_pressure_bubble)

            if hasattr(blk[k], "eq_pressure_dew"):
                calculate_variable_from_constraint(blk[k].pressure_dew,
                                                   blk[k].eq_pressure_dew)

        if outlvl > 0:
            _log.info("Dew and bubble points initialization for "
                      "{} completed".format(blk.name))

        # ---------------------------------------------------------------------
        # If flash, initialize T1 and Teq
        for k in blk.keys():
            if blk[k].config.has_phase_equilibrium:
                blk[k]._t1.value = max(blk[k].temperature.value,
                                       blk[k].temperature_bubble.value)
                blk[k]._teq.value = min(blk[k]._t1.value,
                                        blk[k].temperature_dew.value)

        if outlvl > 0:
            _log.info("Equilibrium temperature initialization for "
                      "{} completed".format(blk.name))

        # ---------------------------------------------------------------------
        # Initialize flow rates and compositions
        # TODO : This will need ot be generalised more when we move to a
        # modular implementation
        for k in blk.keys():
            if blk[k]._params.config.valid_phase == "Liq":
                blk[k].flow_mol_phase['Liq'].value = \
                    blk[k].flow_mol.value

                for j in blk[k]._params.component_list:
                    blk[k].mole_frac_phase_comp['Liq', j].value = \
                        blk[k].mole_frac_comp[j].value

            elif blk[k]._params.config.valid_phase == "Vap":
                blk[k].flow_mol_phase['Vap'].value = \
                    blk[k].flow_mol.value

                for j in blk[k]._params.component_list:
                    blk[k].mole_frac_phase_comp['Vap', j].value = \
                        blk[k].mole_frac_comp[j].value

            else:
                # Seems to work best with default values for phase flows
                for j in blk[k]._params.component_list:
                    blk[k].mole_frac_phase_comp['Vap', j].value = \
                        blk[k].mole_frac_comp[j].value
                    blk[k].mole_frac_phase_comp['Liq', j].value = \
                        blk[k].mole_frac_comp[j].value

                    calculate_variable_from_constraint(
                            blk[k].pressure_sat[j],
                            blk[k].eq_pressure_sat[j])

        # ---------------------------------------------------------------------
        # Solve phase equilibrium constraints
        for k in blk.keys():
            for c in blk[k].component_objects(Constraint):
                # Deactivate all property constraints
                if c.local_name not in ("total_flow_balance",
                                        "component_flow_balances",
                                        "equilibrium_constraint",
                                        "sum_mole_frac",
                                        "_t1_constraint",
                                        "_teq_constraint",
                                        "eq_pressure_dew",
                                        "eq_pressure_bubble",
                                        "eq_temperature_dew",
                                        "eq_temperature_bubble",
                                        "eq_pressure_sat"):
                    c.deactivate()

        results = solve_indexed_blocks(opt, [blk], tee=stee)

        if outlvl > 0:
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                _log.info("Phase state initialization for "
                          "{} completed".format(blk.name))
            else:
                _log.warning("Phase state initialization for "
                             "{} failed".format(blk.name))

        # ---------------------------------------------------------------------
        # Initialize other properties
        for k in blk.keys():
            for c in blk[k].component_objects(Constraint):
                # Activate all constraints except sum_mole_frac_out
                if c.local_name not in ("sum_mole_frac_out"):
                    c.activate()

        if outlvl > 0:
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                _log.info("Property initialization for "
                          "{} completed".format(blk.name))
            else:
                _log.warning("Property initialization for "
                             "{} failed".format(blk.name))

        # ---------------------------------------------------------------------
        # Return state to initial conditions
        for k in blk.keys():
            if (blk[k].config.defined_state is False):
                blk[k].sum_mole_frac_out.activate()

        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                blk.release_state(flags)

        if outlvl > 0:
            _log.info("Initialization completed for {}".format(blk.name))

    def release_state(blk, flags, outlvl=0):
        '''
        Method to relase state variables fixed during initialization.
        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        '''
        if flags is None:
            return

        # Unfix state variables
        revert_state_vars(blk, flags)

        if outlvl > 0:
            if outlvl > 0:
                _log.info('{} states released.'.format(blk.name))

@declare_process_block_class("IdealStateBlock",
                             block_class=_IdealStateBlock)
class IdealStateBlockData(StateBlockData):
    """An example property package for ideal VLE."""

    def build(self):
        """Callable method for Block construction."""
        super(IdealStateBlockData, self).build()

        # Check for valid phase indicator and consistent flags
        if self.config.has_phase_equilibrium and \
                self._params.config.valid_phase in ['Vap', 'Liq']:
            raise ConfigurationError("Inconsistent inputs. Valid phase"
                                     " flag not set to VL for the state"
                                     " block but has_phase_equilibrium"
                                     " is set to True.")

        # Add state variables
        self.flow_mol = Var(initialize=1.0,
                            domain=NonNegativeReals,
                            doc='Component molar flowrate [mol/s]')
        self.mole_frac_comp = Var(
                self._params.component_list,
                bounds=(0, 1),
                initialize=1 / len(self._params.component_list),
                doc='Mixture mole fractions [-]')
        self.pressure = Var(initialize=101325,
                            domain=NonNegativeReals,
                            doc='State pressure [Pa]')
        self.temperature = Var(initialize=298.15,
                               domain=NonNegativeReals,
                               doc='State temperature [K]')

        # Add supporting variables
        self.flow_mol_phase = Var(self._params.phase_list,
                                  initialize=0.5,
                                  doc='Phase molar flow rates [mol/s]')

        self.mole_frac_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            initialize=1 / len(self._params.component_list),
            bounds=(0, 1),
            doc='Phase mole fractions [-]')

        if not self.config.has_phase_equilibrium and \
                self._params.config.valid_phase == "Liq":
            self._make_liq_phase_eq()
        elif not self.config.has_phase_equilibrium and \
                self._params.config.valid_phase == "Vap":
            self._make_vap_phase_eq()
        elif (self.config.has_phase_equilibrium) or \
                (self._params.config.valid_phase ==
                    ('Liq', 'Vap')) or \
                (self._params.config.valid_phase ==
                    ('Vap', 'Liq')):
            self._make_flash_eq()
        else:
            raise BurntToast("{} found unexpected combination of valid_phases "
                             "and has_phase_equilibrium. Please contact the "
                             "IDAES developers with this bug."
                             .format(self.name))

    def _make_liq_phase_eq(self):
        def rule_total_mass_balance(b):
            return b.flow_mol_phase['Liq'] == b.flow_mol
        self.total_flow_balance = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(b, i):
            return b.mole_frac_comp[i] == b.mole_frac_phase_comp['Liq', i]
        self.component_flow_balances = Constraint(self._params.component_list,
                                                  rule=rule_comp_mass_balance)

        if self.config.defined_state is False:
            # applied at outlet only
            self.sum_mole_frac_out = Constraint(
                expr=1 == sum(self.mole_frac_comp[i]
                              for i in self._params.component_list))

    def _make_vap_phase_eq(self):
        def rule_total_mass_balance(b):
            return b.flow_mol_phase['Vap'] == b.flow_mol
        self.total_flow_balance = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(b, i):
            return b.mole_frac_comp[i] == b.mole_frac_phase_comp['Vap', i]
        self.component_flow_balances = Constraint(self._params.component_list,
                                                  rule=rule_comp_mass_balance)

        if self.config.defined_state is False:
            # applied at outlet only
            self.sum_mole_frac_out = \
                Constraint(expr=1 == sum(self.mole_frac_comp[i]
                           for i in self._params.component_list))

    def _make_flash_eq(self):

        def rule_total_mass_balance(b):
            return b.flow_mol_phase['Liq'] + \
                b.flow_mol_phase['Vap'] == b.flow_mol
        self.total_flow_balance = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(b, i):
            return b.flow_mol * b.mole_frac_comp[i] == \
                b.flow_mol_phase['Liq'] * b.mole_frac_phase_comp['Liq', i] + \
                b.flow_mol_phase['Vap'] * b.mole_frac_phase_comp['Vap', i]
        self.component_flow_balances = Constraint(self._params.component_list,
                                                  rule=rule_comp_mass_balance)

        def rule_mole_frac(b):
            return sum(b.mole_frac_phase_comp['Liq', i]
                       for i in b._params.component_list) -\
                sum(b.mole_frac_phase_comp['Vap', i]
                    for i in b._params.component_list) == 0
        self.sum_mole_frac = Constraint(rule=rule_mole_frac)

        if self.config.defined_state is False:
            # applied at outlet only
            self.sum_mole_frac_out = \
                Constraint(expr=1 == sum(self.mole_frac_comp[i]
                           for i in self._params.component_list))

        if self.config.has_phase_equilibrium:
            # Definition of equilibrium temperature for smooth VLE
            self._teq = Var(initialize=self.temperature.value,
                            doc='Temperature for calculating '
                                'phase equilibrium')
            self._t1 = Var(initialize=self.temperature.value,
                           doc='Intermediate temperature for calculating Teq')

            self.eps_1 = Param(default=0.01,
                               mutable=True,
                               doc='Smoothing parameter for Teq')
            self.eps_2 = Param(default=0.0005,
                               mutable=True,
                               doc='Smoothing parameter for Teq')

            # PSE paper Eqn 13
            def rule_t1(b):
                return b._t1 == 0.5 * \
                    (b.temperature + b.temperature_bubble +
                     sqrt((b.temperature - b.temperature_bubble)**2 +
                          b.eps_1**2))
            self._t1_constraint = Constraint(rule=rule_t1)

            # PSE paper Eqn 14
            # TODO : Add option for supercritical extension
            def rule_teq(b):
                return b._teq == 0.5 * (b._t1 + b.temperature_dew -
                                        sqrt((b._t1 - b.temperature_dew)**2 +
                                             b.eps_2**2))
            self._teq_constraint = Constraint(rule=rule_teq)

            def rule_tr_eq(b, i):
                return b._teq / b._params.temperature_crit[i]
            self._tr_eq = Expression(self._params.component_list,
                                     rule=rule_tr_eq,
                                     doc='Component reduced temperatures [-]')

            def rule_equilibrium(b, i):
                return b.fug_vap[i] == b.fug_liq[i]
            self.equilibrium_constraint = \
                Constraint(self._params.component_list, rule=rule_equilibrium)

# -----------------------------------------------------------------------------
# Property Methods
    def _dens_mol_phase(self):
        self.dens_mol_phase = Var(self._params.phase_list,
                                  doc="Molar density [mol/m^3]")

        def rule_dens_mol_phase(b, p):
            if p == 'Vap':
                return b._dens_mol_vap()
            else:
                return b._dens_mol_liq()
        self.eq_dens_mol_phase = Constraint(self._params.phase_list,
                                            rule=rule_dens_mol_phase)

    def _energy_internal_mol_phase_comp(self):
        self.energy_internal_mol_phase_comp = Var(
                self._params.phase_list,
                self._params.component_list,
                doc="Phase-component molar specific internal energies [J/mol]")

        def rule_energy_internal_mol_phase_comp(b, p, j):
            if p == 'Vap':
                return b.energy_internal_mol_phase_comp[p, j] == \
                        b.enth_mol_phase_comp[p, j] - \
                        b._params.gas_const*(b.temperature -
                                             b._params.temeprature_ref)
            else:
                return b.energy_internal_mol_phase_comp[p, j] == \
                        b.enth_mol_phase_comp[p, j]
        self.eq_energy_internal_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_energy_internal_mol_phase_comp)

    def _energy_internal_mol_phase(self):
        self.energy_internal_mol_phase = Var(
            self._params.phase_list,
            doc='Phase molar specific internal energies [J/mol]')

        def rule_energy_internal_mol_phase(b, p):
            return b.energy_internal_mol_phase[p] == sum(
                b.energy_internal_mol_phase_comp[p, i] *
                b.mole_frac_phase_comp[p, i]
                for i in b._params.component_list)
        self.eq_energy_internal_mol_phase = Constraint(
                self._params.phase_list,
                rule=rule_energy_internal_mol_phase)

    def _enth_mol_phase_comp(self):
        self.enth_mol_phase_comp = Var(self._params.phase_list,
                                       self._params.component_list,
                                       doc="Phase-component molar specific "
                                           "enthalpies [J/mol]")

        def rule_enth_mol_phase_comp(b, p, j):
            if p == 'Vap':
                return b._enth_mol_comp_vap(j)
            else:
                return b._enth_mol_comp_liq(j)
        self.eq_enth_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_enth_mol_phase_comp)

    def _enth_mol_phase(self):
        self.enth_mol_phase = Var(
            self._params.phase_list,
            doc='Phase molar specific enthalpies [J/mol]')

        def rule_enth_mol_phase(b, p):
            return b.enth_mol_phase[p] == sum(
                b.enth_mol_phase_comp[p, i] *
                b.mole_frac_phase_comp[p, i]
                for i in b._params.component_list)
        self.eq_enth_mol_phase = Constraint(self._params.phase_list,
                                            rule=rule_enth_mol_phase)

    def _entr_mol_phase_comp(self):
        self.entr_mol_phase_comp = Var(
            self._params.phase_list,
            self._params.component_list,
            doc='Phase-component molar specific entropies [J/mol.K]')

        def rule_entr_mol_phase_comp(b, p, j):
            if p == 'Vap':
                return b._entr_mol_comp_vap(j)
            else:
                return b._entr_mol_comp_liq(j)
        self.eq_entr_mol_phase_comp = Constraint(
            self._params.phase_list,
            self._params.component_list,
            rule=rule_entr_mol_phase_comp)

    def _entr_mol_phase(self):
        self.entr_mol_phase = Var(
            self._params.phase_list,
            doc='Phase molar specific enthropies [J/mol.K]')

        def rule_entr_mol_phase(b, p):
            return b.entr_mol_phase[p] == sum(
                b.entr_mol_phase_comp[p, i] *
                b.mole_frac_phase_comp[p, i]
                for i in b._params.component_list)
        self.eq_entr_mol_phase = Constraint(self._params.phase_list,
                                            rule=rule_entr_mol_phase)

# -----------------------------------------------------------------------------
# General Methods
    def get_material_flow_terms(self, p, j):
        """Create material flow terms for control volume."""
        if j in self._params.component_list:
            return self.flow_mol_phase[p] * self.mole_frac_phase_comp[p, j]
        else:
            return 0

    def get_enthalpy_flow_terms(self, p):
        """Create enthalpy flow terms."""
        return self.flow_mol_phase[p] * self.enth_mol_phase[p]

    def get_material_density_terms(self, p, j):
        """Create material density terms."""
        if j in self._params.component_list:
            return self.dens_mol_phase[p] * self.mole_frac_phase_comp[p, j]
        else:
            return 0

    def get_enthalpy_density_terms(self, p):
        """Create enthalpy density terms."""
        return self.dens_mol_phase[p] * self.energy_internal_mol_phase[p]

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def define_state_vars(self):
        """Define state vars."""
        return {"flow_mol": self.flow_mol,
                "mole_frac_comp": self.mole_frac_comp,
                "temperature": self.temperature,
                "pressure": self.pressure}

    def model_check(blk):
        """Model checks for property block."""
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            _log.error('{} Temperature set below lower bound.'
                       .format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            _log.error('{} Temperature set above upper bound.'
                       .format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            _log.error('{} Pressure set below lower bound.'.format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            _log.error('{} Pressure set above upper bound.'.format(blk.name))

    # Property package utility functions
    def calculate_bubble_point_temperature(self, clear_components=True):
        """"To compute the bubble point temperature of the mixture."""

        if hasattr(self, "eq_temperature_bubble"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(self.temperature_bubble,
                                           self.eq_temperature_bubble)

        return self.temperature_bubble.value

        if clear_components is True:
            self.del_component(self.eq_temperature_bubble)
            self.del_component(self._p_sat_bubbleT)
            self.del_component(self.temperature_bubble)

    def calculate_dew_point_temperature(self, clear_components=True):
        """"To compute the dew point temperature of the mixture."""

        if hasattr(self, "eq_temperature_dew"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(self.temperature_dew,
                                           self.eq_temperature_dew)

        return self.temperature_dew.value

        # Delete the var/constraint created in this method that are part of the
        # IdealStateBlock if the user desires
        if clear_components is True:
            self.del_component(self.eq_temperature_dew)
            self.del_component(self._p_sat_dewT)
            self.del_component(self.temperature_dew)

    def calculate_bubble_point_pressure(self, clear_components=True):
        """"To compute the bubble point pressure of the mixture."""

        if hasattr(self, "eq_pressure_bubble"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(self.pressure_bubble,
                                           self.eq_pressure_bubble)

        return self.pressure_bubble.value

        # Delete the var/constraint created in this method that are part of the
        # IdealStateBlock if the user desires
        if clear_components is True:
            self.del_component(self.eq_pressure_bubble)
            self.del_component(self._p_sat_bubbleP)
            self.del_component(self.pressure_bubble)

    def calculate_dew_point_pressure(self, clear_components=True):
        """"To compute the dew point pressure of the mixture."""

        if hasattr(self, "eq_pressure_dew"):
            # Do not delete components if the block already has the components
            clear_components = False

        calculate_variable_from_constraint(self.pressure_dew,
                                           self.eq_pressure_dew)

        return self.pressure_dew.value

        # Delete the var/constraint created in this method that are part of the
        # IdealStateBlock if the user desires
        if clear_components is True:
            self.del_component(self.eq_pressure_dew)
            self.del_component(self._p_sat_dewP)
            self.del_component(self.pressure_dew)

# -----------------------------------------------------------------------------
# Bubble and Dew Points
# Ideal-Ideal properties allow for the simplifications below
# Other methods require more complex equations with shadow compositions

# For future work, propose the following:
# Core class writes a set of constraints Phi_L_i == Phi_V_i
# Phi_L_i and Phi_V_i make calls to submethods which add shadow compositions
# as needed
    def _temperature_bubble(self):
        self.temperature_bubble = Var(initialize=298.15,
                                      doc="Bubble point temperature (K)")

        def rule_psat_bubble(b, j):
            return 1e5*10**(b._params.pressure_sat_coeff[j, 'A'] -
                            b._params.pressure_sat_coeff[j, 'B'] /
                            (b.temperature_bubble +
                             b._params.pressure_sat_coeff[j, 'C']))
        try:
            # Try to build expression
            self._p_sat_bubbleT = Expression(self._params.component_list,
                                             rule=rule_psat_bubble)

            def rule_temp_bubble(b):
                return sum(b._p_sat_bubbleT[i] * b.mole_frac_comp[i]
                           for i in b._params.component_list) - \
                    b.pressure == 0
            self.eq_temperature_bubble = Constraint(rule=rule_temp_bubble)

        except AttributeError:
            # If expression fails, clean up so that DAE can try again later
            # Deleting only var/expression as expression construction will fail
            # first; if it passes then constraint construction will not fail.
            self.del_component(self.temperature_bubble)
            self.del_component(self._p_sat_bubbleT)

    def _temperature_dew(self):

        self.temperature_dew = Var(initialize=298.15,
                                   doc="Dew point temperature (K)")

        def rule_psat_dew(b, j):
            return 1e5*10**(b._params.pressure_sat_coeff[j, 'A'] -
                            b._params.pressure_sat_coeff[j, 'B'] /
                            (b.temperature_dew +
                             b._params.pressure_sat_coeff[j, 'C']))

        try:
            # Try to build expression
            self._p_sat_dewT = Expression(self._params.component_list,
                                          rule=rule_psat_dew)

            def rule_temp_dew(b):
                return b.pressure * sum(b.mole_frac_comp[i] /
                                        b._p_sat_dewT[i]
                                        for i in b._params.component_list) \
                    - 1 == 0
            self.eq_temperature_dew = Constraint(rule=rule_temp_dew)
        except AttributeError:
            # If expression fails, clean up so that DAE can try again later
            # Deleting only var/expression as expression construction will fail
            # first; if it passes then constraint construction will not fail.
            self.del_component(self.temperature_dew)
            self.del_component(self._p_sat_dewT)

    def _pressure_bubble(self):
        self.pressure_bubble = Var(initialize=298.15,
                                   doc="Bubble point pressure (Pa)")

        def rule_psat_bubble(b, j):
            return 1e5*10**(b._params.pressure_sat_coeff[j, 'A'] -
                            b._params.pressure_sat_coeff[j, 'B'] /
                            (b.temperature +
                             b._params.pressure_sat_coeff[j, 'C']))

        try:
            # Try to build expression
            self._p_sat_bubbleP = Expression(self._params.component_list,
                                             rule=rule_psat_bubble)

            def rule_pressure_bubble(b):
                return sum(b._p_sat_bubbleP[i] * b.mole_frac_comp[i]
                           for i in b._params.component_list) \
                    - b.pressure_bubble == 0
            self.eq_pressure_bubble = Constraint(rule=rule_pressure_bubble)
        except AttributeError:
            # If expression fails, clean up so that DAE can try again later
            # Deleting only var/expression as expression construction will fail
            # first; if it passes then constraint construction will not fail.
            self.del_component(self.pressure_bubble)
            self.del_component(self._p_sat_bubbleP)

    def _pressure_dew(self):
        self.pressure_dew = Var(initialize=298.15,
                                doc="Dew point pressure (Pa)")

        def rule_psat_dew(b, j):
            return 1e5*10**(b._params.pressure_sat_coeff[j, 'A'] -
                            b._params.pressure_sat_coeff[j, 'B'] /
                            (b.temperature +
                             b._params.pressure_sat_coeff[j, 'C']))

        try:
            # Try to build expression
            self._p_sat_dewP = Expression(self._params.component_list,
                                          rule=rule_psat_dew)

            def rule_pressure_dew(b):
                return b.pressure_dew * \
                    sum(b.mole_frac_comp[i] / b._p_sat_dewP[i]
                        for i in b._params.component_list) \
                    - 1 == 0
            self.eq_pressure_dew = Constraint(rule=rule_pressure_dew)
        except AttributeError:
            # If expression fails, clean up so that DAE can try again later
            # Deleting only var/expression as expression construction will fail
            # first; if it passes then constraint construction will not fail.
            self.del_component(self.pressure_dew)
            self.del_component(self._p_sat_dewP)

# -----------------------------------------------------------------------------
# Liquid phase properties
    def _dens_mol_liq(b):
        return b.dens_mol_phase['Liq'] == 1e3 * sum(
            b.mole_frac_phase_comp['Liq', j] *
            b._params.dens_liq_params[j, '1'] /
            b._params.dens_liq_params[j, '2'] **
            (1 + (1 - b.temperature /
                  b._params.dens_liq_params[j, '3']) **
             b._params.dens_liq_params[j, '4'])
            for j in b._params.component_list)

    def _fug_liq(self):
        def fug_liq_rule(b, i):
            return b.pressure_sat[i] * b.mole_frac_phase_comp['Liq', i]
        self.fug_liq = Expression(self._params.component_list,
                                  rule=fug_liq_rule)

    def _pressure_sat(self):
        self.pressure_sat = Var(self._params.component_list,
                                initialize=101325,
                                doc="Vapor pressure [Pa]")

        def rule_P_sat(b, j):
            return ((log10(b.pressure_sat[j]*1e-5) -
                     b._params.pressure_sat_coeff[j, 'A']) *
                    (b._teq + b._params.pressure_sat_coeff[j, 'C'])) == \
                   -b._params.pressure_sat_coeff[j, 'B']
        self.eq_pressure_sat = Constraint(self._params.component_list,
                                          rule=rule_P_sat)

    def _enth_mol_comp_liq(b, j):
        return b.enth_mol_phase_comp['Liq', j] * 1E3 == \
            ((b._params.cp_ig['Liq', j, '5'] / 5) *
                (b.temperature**5 - b._params.temperature_ref**5)
                + (b._params.cp_ig['Liq', j, '4'] / 4) *
                  (b.temperature**4 - b._params.temperature_ref**4)
                + (b._params.cp_ig['Liq', j, '3'] / 3) *
                  (b.temperature**3 - b._params.temperature_ref**3)
                + (b._params.cp_ig['Liq', j, '2'] / 2) *
                  (b.temperature**2 - b._params.temperature_ref**2)
                + b._params.cp_ig['Liq', j, '1'] *
                  (b.temperature - b._params.temperature_ref))

    def _entr_mol_comp_liq(b, j):
        return b.entr_mol_phase_comp['Liq', j] * 1E3 == (
            ((b._params.cp_ig['Liq', j, '5'] / 4) *
                (b.temperature**4 - b._params.temperature_ref**4)
                + (b._params.cp_ig['Liq', j, '4'] / 3) *
                  (b.temperature**3 - b._params.temperature_ref**3)
                + (b._params.cp_ig['Liq', j, '3'] / 2) *
                  (b.temperature**2 - b._params.temperature_ref**2)
                + b._params.cp_ig['Liq', j, '2'] *
                  (b.temperature - b._params.temperature_ref)
                + b._params.cp_ig['Liq', j, '1'] *
                  log(b.temperature / b._params.temperature_ref)) -
            b._params.gas_const *
            log(b.mole_frac_phase_comp['Liq', j] * b.pressure /
                b._params.pressure_ref))

# -----------------------------------------------------------------------------
# Vapour phase properties
    def _dens_mol_vap(b):
        return b.pressure == (b.dens_mol_phase['Vap'] *
                              b._params.gas_const *
                              b.temperature)

    def _fug_vap(self):
        def fug_vap_rule(b, i):
            return b.mole_frac_phase_comp['Vap', i] * b.pressure
        self.fug_vap = Expression(self._params.component_list,
                                  rule=fug_vap_rule)

    def _dh_vap(self):

        # heat of vaporization
        add_object_reference(self, "dh_vap", self._params.dh_vap)

    def _ds_vap(self):
        # entropy of vaporization = dh_Vap/T_boil
        # TODO : something more rigorous would be nice
        self.ds_vap = Var(self._params.component_list,
                          initialize=86,
                          doc="Entropy of vaporization [J/mol.K]")

        def rule_ds_vap(b, j):
            return b.dh_vap[j] == (b.ds_vap[j] *
                                   b._params.temperature_boil[j])
        self.eq_ds_vap = Constraint(self._params.component_list,
                                    rule=rule_ds_vap)

    def _enth_mol_comp_vap(b, j):
        return b.enth_mol_phase_comp['Vap', j] == b.dh_vap[j] + \
            ((b._params.cp_ig['Vap', j, '5'] / 5) *
             (b.temperature**5 - b._params.temperature_ref**5)
             + (b._params.cp_ig['Vap', j, '4'] / 4) *
               (b.temperature**4 - b._params.temperature_ref**4)
             + (b._params.cp_ig['Vap', j, '3'] / 3) *
               (b.temperature**3 - b._params.temperature_ref**3)
             + (b._params.cp_ig['Vap', j, '2'] / 2) *
               (b.temperature**2 - b._params.temperature_ref**2)
             + b._params.cp_ig['Vap', j, '1'] *
               (b.temperature - b._params.temperature_ref))

    def _entr_mol_comp_vap(b, j):
        return b.entr_mol_phase_comp['Vap', j] == (
            b.ds_vap[j] + ((b._params.cp_ig['Vap', j, '5'] / 4) *
                           (b.temperature**4 - b._params.temperature_ref**4)
                           + (b._params.cp_ig['Vap', j, '4'] / 3) *
                           (b.temperature**3 - b._params.temperature_ref**3)
                           + (b._params.cp_ig['Vap', j, '3'] / 2) *
                           (b.temperature**2 - b._params.temperature_ref**2)
                           + b._params.cp_ig['Vap', j, '2'] *
                           (b.temperature - b._params.temperature_ref)
                           + b._params.cp_ig['Vap', j, '1'] *
                           log(b.temperature / b._params.temperature_ref)) -
            b._params.gas_const * log(b.mole_frac_phase_comp['Vap', j] * b.pressure /
                                      b._params.pressure_ref))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for unit_model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import Block, ConcreteModel, Set, Var
from pyomo.network import Port

from idaes.core import (FlowsheetBlockData, declare_process_block_class,
                        UnitModelBlockData, useDefault, PhysicalParameterBlock,
                        StateBlock, StateBlockData, ControlVolume0DBlock)
from idaes.core.util.exceptions import ConfigurationError, DynamicError


@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("PhysicalParameterTestBlock")
class _PhysicalParameterBlock(PhysicalParameterBlock):
    def build(self):
        super(_PhysicalParameterBlock, self).build()

        self.phase_list = Set(initialize=["p1", "p2"])
        self.component_list = Set(initialize=["c1", "c2"])

        self.state_block_class = TestStateBlock


@declare_process_block_class("TestStateBlock", block_class=StateBlock)
class StateTestBlockData(StateBlockData):
    def build(self):
        super(StateTestBlockData, self).build()

        self.a = Var(initialize=1)
        self.b = Var(initialize=2)
        self.c = Var(initialize=3)

    def define_state_vars(self):
        return {"a": self.a,
                "b": self.b,
                "c": self.c}


@declare_process_block_class("Unit")
class UnitData(UnitModelBlockData):
    def build(self):
        super(UnitModelBlockData, self).build()


def test_config_block():
    m = ConcreteModel()

    m.u = Unit()

    assert len(m.u. config) == 2
    assert m.u.config.dynamic == useDefault


def test_config_args():
    m = ConcreteModel()

    m.u = Unit(default={"dynamic": True})

    assert m.u.config.dynamic is True


def test_config_args_invalid():
    # Test validation of config arguments
    m = ConcreteModel()

    m.u = Unit()

    m.u.config.dynamic = True
    m.u.config.dynamic = False
    m.u.config.dynamic = None

    # Test that Value error raised when given invalid config arguments
    with pytest.raises(ValueError):
        m.u.config.dynamic = "foo"  # invalid str
    with pytest.raises(ValueError):
        m.u.config.dynamic = 5  # invalid int
    with pytest.raises(ValueError):
        m.u.config.dynamic = 2.0  # invalid float
    with pytest.raises(ValueError):
        m.u.config.dynamic = [2.0]  # invalid list
    with pytest.raises(ValueError):
        m.u.config.dynamic = {'a': 2.0}  # invalid dict


def test_setup_dynamics1():
    # Test that _setup_dynamics gets argument from parent
    m = ConcreteModel()

    m.fs = Flowsheet(default={"dynamic": False})

    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    assert m.fs.u.config.dynamic is False


def test_setup_dynamics2():
    # Test that _setup_dynamics returns an DynamicError when parent has no
    # dynamic config argument

    m = ConcreteModel()
    # Intermediate Block is required, as ConcreteModels have a config attribute
    m.b = Block()
    m.b.u = Unit()

    with pytest.raises(DynamicError):
        m.b.u._setup_dynamics()


def test_setup_dynamics_dynamic_in_steady_state():
    # Test that a DynamicError is raised when a dynamic models is placed in a
    # steady-state parent
    m = ConcreteModel()

    m.fs = Flowsheet(default={"dynamic": False})

    m.fs.u = Unit(default={"dynamic": True})
    with pytest.raises(DynamicError):
        m.fs.u._setup_dynamics()


def test_setup_dynamics_get_time():
    # Test that time domain is collected correctly
    m = ConcreteModel()

    m.fs = Flowsheet(default={"dynamic": False})

    m.fs.u = Unit()
    m.fs.u._setup_dynamics()


def test_setup_dynamics_has_holdup():
    # Test that has_holdup argument is True when dynamic is True
    m = ConcreteModel()

    m.fs = Flowsheet(default={"dynamic": True})

    m.fs.u = Unit()
    m.fs.u.config.has_holdup=False

    with pytest.raises(ConfigurationError):
        m.fs.u._setup_dynamics()


def test_add_port():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.prop = TestStateBlock(m.fs.time,
                             default={"parameters": m.fs.pp})

    p_obj = m.fs.u.add_port(name="test_port", block=m.fs.u.prop)

    assert isinstance(p_obj, Port)
    assert hasattr(m.fs.u, "test_port")
    assert len(m.fs.u.test_port) == 1
    assert m.fs.u.test_port.a[0].value == m.fs.u.prop[0].a.value
    assert m.fs.u.test_port.b[0].value == m.fs.u.prop[0].b.value
    assert m.fs.u.test_port.c[0].value == m.fs.u.prop[0].c.value


def test_add_port_invalid_block():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.prop = TestStateBlock(m.fs.time,
                             default={"parameters": m.fs.pp})

    with pytest.raises(ConfigurationError):
        m.fs.u.add_port(name="test_port", block=m.fs.u)


def test_add_inlet_port_CV0D():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_inlet_port()

    assert isinstance(p_obj, Port)
    assert hasattr(m.fs.u, "inlet")
    assert len(m.fs.u.inlet) == 1

    # Set new inlet conditions to differentiate from outlet
    m.fs.u.control_volume.properties_in[0].a = 10
    m.fs.u.control_volume.properties_in[0].b = 20
    m.fs.u.control_volume.properties_in[0].c = 30

    assert m.fs.u.inlet.a[0].value == \
        m.fs.u.control_volume.properties_in[0].a.value
    assert m.fs.u.inlet.b[0].value == \
        m.fs.u.control_volume.properties_in[0].b.value
    assert m.fs.u.inlet.c[0].value == \
        m.fs.u.control_volume.properties_in[0].c.value


def test_add_inlet_port_CV0D_no_default_block():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    with pytest.raises(ConfigurationError):
        m.fs.u.add_inlet_port()


def test_add_inlet_port_CV0D_full_args():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    m.fs.u.cv.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_inlet_port(name="test_port",
                                  block=m.fs.u.cv,
                                  doc="Test")

    assert isinstance(p_obj, Port)
    assert hasattr(m.fs.u, "test_port")
    assert len(m.fs.u.test_port) == 1

    # Set new inlet conditions to differentiate from outlet
    m.fs.u.cv.properties_in[0].a = 10
    m.fs.u.cv.properties_in[0].b = 20
    m.fs.u.cv.properties_in[0].c = 30

    assert m.fs.u.test_port.a[0].value == \
        m.fs.u.cv.properties_in[0].a.value
    assert m.fs.u.test_port.b[0].value == \
        m.fs.u.cv.properties_in[0].b.value
    assert m.fs.u.test_port.c[0].value == \
        m.fs.u.cv.properties_in[0].c.value


def test_add_inlet_port_CV0D_part_args():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    m.fs.u.cv.add_state_blocks()

    with pytest.raises(ConfigurationError):
        m.fs.u.add_inlet_port(block=m.fs.u.cv)

    with pytest.raises(ConfigurationError):
        m.fs.u.add_inlet_port(name="foo")


def test_add_outlet_port_CV0D():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_outlet_port()

    assert isinstance(p_obj, Port)
    assert hasattr(m.fs.u, "outlet")
    assert len(m.fs.u.outlet) == 1

    # Set new outlet conditions to differentiate from intlet
    m.fs.u.control_volume.properties_out[0].a = 10
    m.fs.u.control_volume.properties_out[0].b = 20
    m.fs.u.control_volume.properties_out[0].c = 30

    assert m.fs.u.outlet.a[0].value == \
        m.fs.u.control_volume.properties_out[0].a.value
    assert m.fs.u.outlet.b[0].value == \
        m.fs.u.control_volume.properties_out[0].b.value
    assert m.fs.u.outlet.c[0].value == \
        m.fs.u.control_volume.properties_out[0].c.value


def test_add_outlet_port_CV0D_no_default_block():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    with pytest.raises(ConfigurationError):
        m.fs.u.add_outlet_port()


def test_add_outlet_port_CV0D_full_args():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    m.fs.u.cv.add_state_blocks(has_phase_equilibrium=False)

    p_obj = m.fs.u.add_outlet_port(name="test_port",
                                  block=m.fs.u.cv,
                                  doc="Test")

    assert isinstance(p_obj, Port)
    assert hasattr(m.fs.u, "test_port")
    assert len(m.fs.u.test_port) == 1

    # Set new outlet conditions to differentiate from inlet
    m.fs.u.cv.properties_out[0].a = 10
    m.fs.u.cv.properties_out[0].b = 20
    m.fs.u.cv.properties_out[0].c = 30

    assert m.fs.u.test_port.a[0].value == \
        m.fs.u.cv.properties_out[0].a.value
    assert m.fs.u.test_port.b[0].value == \
        m.fs.u.cv.properties_out[0].b.value
    assert m.fs.u.test_port.c[0].value == \
        m.fs.u.cv.properties_out[0].c.value


def test_add_inlet_port_CV0D_part_args():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.cv = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    m.fs.u.cv.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.u.add_outlet_port(block=m.fs.u.cv)

    with pytest.raises(ConfigurationError):
        m.fs.u.add_outlet_port(name="foo")


def test_fix_unfix_initial_conditions():
    fs = Flowsheet(default={"dynamic": True, "time_set": [0, 1, 2]},
                   concrete=True)
    fs._setup_dynamics()

    fs.b = Unit()

    fs.b.material_accumulation = Var(fs.time, ["a", "b", "c"])
    fs.b.element_accumulation = Var(fs.time, ["a", "b", "c"])
    fs.b.enthalpy_accumulation = Var(fs.time)

    fs.fix_initial_conditions()

    for t in fs.time:
        for j in ["a", "b", "c"]:
            if t == 0:
                assert fs.b.material_accumulation[t, j].fixed
                assert fs.b.element_accumulation[t, j].fixed
                assert fs.b.enthalpy_accumulation[t].fixed
            else:
                assert fs.b.material_accumulation[t, j].fixed is False
                assert fs.b.element_accumulation[t, j].fixed is False
                assert fs.b.enthalpy_accumulation[t].fixed is False

    fs.unfix_initial_conditions()

    for t in fs.time:
        for j in ["a", "b", "c"]:
            assert fs.b.material_accumulation[t, j].fixed is False
            assert fs.b.element_accumulation[t, j].fixed is False
            assert fs.b.enthalpy_accumulation[t].fixed is False


def test_get_stream_table_contents_CV0D():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    m.fs.u.add_inlet_port()
    m.fs.u.add_outlet_port()

    df = m.fs.u._get_stream_table_contents()

    assert df.loc["a"]["Inlet"] == 1
    assert df.loc["b"]["Inlet"] == 2
    assert df.loc["c"]["Inlet"] == 3

    assert df.loc["a"]["Outlet"] == 1
    assert df.loc["b"]["Outlet"] == 2
    assert df.loc["c"]["Outlet"] == 3


def test_get_stream_table_contents_CV0D_missing_default_port():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    with pytest.raises(ConfigurationError):
        m.fs.u._get_stream_table_contents()


def test_report_CV0D():
    m = ConcreteModel()
    m.fs = Flowsheet()
    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.u = Unit()
    m.fs.u._setup_dynamics()

    m.fs.u.control_volume = ControlVolume0DBlock(
            default={"property_package": m.fs.pp})

    m.fs.u.control_volume.add_state_blocks(has_phase_equilibrium=False)

    m.fs.u.add_inlet_port()
    m.fs.u.add_outlet_port()

    m.fs.u.report()



import logging
import os


def level_num(name, default_level):
    """Get level num from name"""
    level_mapping = dict(ERROR=logging.ERROR, INFO=logging.INFO,
                         WARNING=logging.WARNING, DEBUG=logging.DEBUG)
    return level_mapping.get(name.upper(), default_level)


# set up test logger
_log = logging.getLogger(__name__)

level = logging.WARNING
if 'IDAES_TEST_LOG_LEVEL' in os.environ:
    env_level = os.environ['IDAES_TEST_LOG_LEVEL']
    level = level_num(env_level, level)

_h = logging.StreamHandler()
_h.setFormatter(logging.Formatter('%(asctime)s [%(levelname)s] '
                                  '%(filename)s:%(lineno)d :: %(message)s'))
_log.addHandler(_h)
_log.setLevel(level)
_log.propagate = False